from functools import cached_property
from typing import TYPE_CHECKING, List, cast

import boto3

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.services import (
        CacheSecurityGroupMembership,
    )


class ElastiCacheManagerTagsMixin:
    """
    Used on both :py:class:`botocraft.services.elasticache.CacheCluster` and
    :py:class:`botocraft.services.elasticache.ReplicationGroup` to implement the
    ``tags`` relation.
    """

    def get_tags(self, arn: str) -> dict[str, str]:
        """
        Get the tags for the elasticache resource identified by ``arn``.

        Args:
            arn: The ARN of the elasticache resource

        Returns:
            A dictionary of key/value pairs, where the key is the tag name and
            the value is the tag value

        """
        tags = self.client.list_tags_for_resource(ResourceName=arn)["TagList"]  # type: ignore[attr-defined]
        # Convert the list of tags to a dictionary
        _tags: dict[str, str] = {}
        for tag in tags:
            _tags[tag["Key"]] = tag["Value"]
        return _tags


class CacheClusterModelMixin:
    """
    A mixin is used on :py:class:`botocraft.services.elasticache.CacheCluster`
    implement the "security_groups" relation.   Normally we would use a
    "relation" type in the model definition to use the .list() function to list
    what we want, but ``describe_cache_clusters`` either lists a single cluster
    or all clusters, so we need to roll our own method
    """

    session: boto3.session.Session
    CacheSecurityGroups: List["CacheSecurityGroupMembership"]

    @property
    def security_groups(self) -> "PrimaryBoto3ModelQuerySet":
        """
        List all the :py:class:`CacheCluster` objects that are part of this
        replication group.
        """
        # We have to do the actual import here to avoid circular imports
        from botocraft.services import CacheSecurityGroup

        names = [x.CacheSecurityGroupName for x in self.CacheSecurityGroups]
        return PrimaryBoto3ModelQuerySet(
            [
                CacheSecurityGroup.objects.using(self.session).get(group_name)
                for group_name in names
            ]
        )  # type: ignore[arg-type]

    @cached_property
    def hostname(self) -> str:
        """
        The hostname of the cache cluster.

        Note:
            This is the hostname of the first node in the cluster.  If you have
            a cluster with multiple nodes, you will need to use the ``CacheNodes``
            property to get the hostnames of the other nodes.

        """
        return cast("str", self.CacheNodes[0].Endpoint.Address)  # type: ignore[attr-defined]

    @cached_property
    def port(self) -> int:
        """
        The port of the cache cluster.

        Note:
            This is the port of the first node in the cluster.  If you have
            a cluster with multiple nodes, you will need to use the ``CacheNodes``
            property to get the ports of the other nodes.

        """
        return cast("int", self.CacheNodes[0].Endpoint.Port)  # type: ignore[attr-defined]

    @cached_property
    def tags(self) -> dict[str, str]:
        """
        Get the tags for the cache cluster.

        This is a dictionary of key/value pairs, where the key is the tag name
        and the value is the tag value
        """
        return self.objects.using(self.session).get_tags(self.arn)  # type: ignore[attr-defined]


class ReplicationGroupModelMixin:
    """
    A mixin is used on :py:class:`botocraft.services.elasticache.ReplicationGroup`
    implement the "clusters" relation.   Normally we would use a "relation" type
    in the model definition to use the .list() function to list what we want, but
    ``describe_cache_clusters`` either lists a single cluster or all clusters,
    so we need to roll our own method
    """

    session: boto3.session.Session
    MemberClusters: List[str]

    @cached_property
    def clusters(self) -> "PrimaryBoto3ModelQuerySet":
        """
        List all the :py:class:`CacheCluster` objects that are part of this
        replication group.
        """
        # We have to do the actual import here to avoid circular imports
        from botocraft.services import CacheCluster

        return PrimaryBoto3ModelQuerySet(
            [
                CacheCluster.objects.using(self.session).get(cluster_id)
                for cluster_id in self.MemberClusters
            ]
        )  # type: ignore[arg-type]

    @property
    def engine_version(self) -> str:
        """
        The engine version of the replication group.  This is the same as the
        engine version of the first cluster in the replication group.
        """
        from botocraft.services import CacheCluster

        cluster = CacheCluster.objects.using(self.session).get(
            self.MemberClusters[0], ShowCacheNodeInfo=False
        )
        return cast("str", cluster.EngineVersion)

    @cached_property
    def hostname(self) -> str:
        """
        The hostname of the cache cluster.
        """
        return cast("str", self.NodeGroups[0].PrimaryEndpoint.Address)  # type: ignore[attr-defined]

    @cached_property
    def port(self) -> int:
        """
        The port of the cache cluster.
        """
        return cast("int", self.NodeGroups[0].PrimaryEndpoint.Port)  # type: ignore[attr-defined]

    @cached_property
    def tags(self) -> dict[str, str]:
        """
        Get the tags for the replication group.

        This is a dictionary of key/value pairs, where the key is the tag name
        and the value is the tag value
        """
        return self.objects.using(self.session).get_tags(self.arn)  # type: ignore[attr-defined]



from pathlib import Path
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from botocraft.sync.service import ServiceGenerator


class ServiceSphinxDocBuilder:
    """
    A Sphinx documentation builder for a single service.

    Do the following things:

      * Gather the managers and models from the service generator.
      * Segregate the models into primary and secondary models
      * Create a Sphinx model page for the service, containing the managers,
        primary models, and secondary models.
      * Write the model page to ``../../doc/source/api/services/<service_name>.rst``
      * Return the path to that file, relative to ``../../doc/source/``

    The service page in the docs should end up looking something like this:

    .. code-block:: rst

        Elastic Container Service (ecs)
        ===============================

        Primary Models
        --------------

        .. autopydantic_model:: botocraft.services.ecs.Cluster

        .. autopydantic_model:: botocraft.services.ecs.Service

        Managers
        --------

        .. autoclass:: botocraft.services.ecs.ClusterManager
            :members:
            :show-inheritance:

        .. autoclass:: botocraft.services.ecs.ServiceManager
            :members:
            :show-inheritance:

        Secondary Models
        ----------------

        .. autopydantic_model:: botocraft.services.ecs.ClusterSetting

        Request/Response Models
        -----------------------

        .. autopydantic_model:: botocraft.services.ecs.ClusterSetting


        ...


    Args:
        generator: The service generator.

    """

    def __init__(self, generator: "ServiceGenerator") -> None:
        #: The service generator.
        self.generator = generator

    @property
    def path(self) -> Path:
        """
        The path to the service documentation file.
        """
        # First get the path to this file
        this_file = Path(__file__).resolve()
        docs_dir = this_file.parents[2] / "doc" / "source" / "api" / "services"
        return docs_dir / f"{self.generator.safe_service_name}.rst"

    @property
    def header(self) -> str:
        """
        The header for the service documentation.
        """
        title = (
            f"{self.generator.service_full_name} ({self.generator.aws_service_name})"
        )
        return f"""
{title}
{"=" * len(title)}
"""

    def autoclass(self, class_name: str, pydantic: bool = True) -> str:
        """
        The Sphinx documentation for a manager.

        Args:
            class_name: The name of the class to document
            pydantic: Whether to use autopydantic_model or autoclass

        """
        if pydantic:
            return f"""

.. autopydantic_model:: botocraft.services.{self.generator.safe_service_name}.{class_name}
    :show-inheritance:
    :inherited-members:
    :exclude-members: update_forward_refs, model_extra, model_fields_set, validate, schema_json, model_rebuild, model_post_init, model_parametrized_name, model_json_schema, copy, from_orm, dict, json, schema, schema_json, model_dump, construct, model_copy, model_validate, model_validate_json, model_validate_dict, model_validate_json_schema, model_validate_python, model_dump_json, model_dump_json_schema, model_dump_dict, parse_file, parse_obj, parse_raw, parse_json, parse_file_json, parse_file_dict, parse_file_json_schema, parse_file_python
"""  # noqa: E501
        return f"""
.. autoclass:: botocraft.services.{self.generator.safe_service_name}.{class_name}
   :members:
   :show-inheritance:
"""

    def classes(self, names: List[str], pydantic: bool = True) -> str:
        """
        The Sphinx documentation for the managers.
        """
        names.sort()
        code = ""
        for name in names:
            code += self.autoclass(name, pydantic=pydantic)
        return f"""{code}
"""

    @property
    def body(self) -> str:
        """
        The body of the documentation. for the service.
        """
        managers = list(self.generator.manager_classes.keys())
        managers_doc = self.classes(managers, pydantic=False)
        primary_models = [
            model
            for model in self.generator.model_classes
            if model in self.generator.service_def.primary_models
        ]
        alternate_names = [
            model.alternate_name
            for model in self.generator.service_def.primary_models.values()
            if model.alternate_name
        ]
        # Extend the primary models with those primary models that have
        # alternate names
        primary_models.extend(
            [
                model_name
                for model_name in self.generator.model_classes
                if model_name in alternate_names
            ]
        )
        primary_models.sort()
        # Ensure that the models are unique
        primary_models = list(set(primary_models))
        primary_models_doc = self.classes(primary_models)
        secondary_models = []
        for model in self.generator.model_classes:
            if model in primary_models:
                continue
            secondary_models.append(model)
        secondary_models.sort()
        # Ensure that the models are unique
        secondary_models = list(set(secondary_models))
        secondary_models_doc = self.classes(secondary_models)
        response_models = list(self.generator.response_classes.keys())
        response_models_doc = self.classes(response_models)
        return f"""
{self.header}

Primary Models
--------------

Primary models are models that you can act on directly. They are the models that
represent resources in the AWS service, and are acted on by the managers.

{primary_models_doc}

Managers
--------

Managers work with the primary models to provide a high-level interface to the
AWS service. They are responsible for creating, updating, and deleting the
resources in the service, as well as any additional operations that are
available for those models.

{managers_doc}


Secondary Models
----------------

Secondary models are models that are used by the primary models to organize
their data. They are not acted on directly, but are used to describe the
structure of the fields in the primary models or other secondary models.

{secondary_models_doc}

Request/Response Models
-----------------------

Request/response models are models that are used to describe the structure of
the data that is sent to and received from the AWS service. They are used by
the managers to send requests to the service and to parse the responses that
are received.

You will not often use them directly -- typically they are used by the managers
internally to send requests and parse responses -- but they are included here
for completeness, and because occasionally we return them directly to you
because they have some useful additional information.

{response_models_doc}
"""

    def write(self) -> None:
        """
        Write the generated docs to the output file, and format it with black,
        isort, and docformatter.

        Args:
            code: the code to write to the output file.

        """
        # First format the code with black so we fix most of the formatting
        # issues.
        with self.path.open("w", encoding="utf-8") as fd:
            fd.write(self.body)



# This file is automatically generated by botocraft.  Do not edit directly.
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
from botocraft.mixins.elbv2 import load_balancer_attributes_to_dict
from botocraft.mixins.tags import TagsDictMixin
from botocraft.services.common import Tag
from .abstract import (
    Boto3Model,
    ReadonlyBoto3Model,
    PrimaryBoto3Model,
    ReadonlyPrimaryBoto3Model,
    Boto3ModelManager,
    ReadonlyBoto3ModelManager,
)
from botocraft.services.ec2 import Vpc, VpcManager
from typing import ClassVar, Type, Optional, Literal, Dict, List, Union, Any, cast
from botocraft.services.ec2 import SecurityGroup, SecurityGroupManager
from functools import cached_property
from .abstract import PrimaryBoto3ModelQuerySet
from datetime import datetime
from collections import OrderedDict
from pydantic import Field

# ===============
# Managers
# ===============


class LoadBalancerManager(Boto3ModelManager):
    service_name: str = "elbv2"

    def create(
        self,
        model: "LoadBalancer",
        SubnetMappings: List["SubnetMapping"] | None = None,
        Tags: List[Tag] | None = None,
    ) -> "LoadBalancer":
        """
        Creates an Application Load Balancer, Network Load Balancer, or Gateway Load
        Balancer.

        Args:
            model: The :py:class:`LoadBalancer` to create.

        Keyword Args:
            SubnetMappings: The IDs of the subnets. You can specify only one subnet per Availability Zone. You must specify
                either subnets or subnet mappings, but not both.
            Tags: The tags to assign to the load balancer.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Name=data.get("LoadBalancerName"),
            Subnets=data.get("Subnets"),
            SubnetMappings=self.serialize(SubnetMappings),
            SecurityGroups=data.get("SecurityGroups"),
            Scheme=data.get("Scheme"),
            Tags=self.serialize(Tags),
            Type=data.get("Type"),
            IpAddressType=data.get("IpAddressType"),
            CustomerOwnedIpv4Pool=data.get("CustomerOwnedIpv4Pool"),
            EnablePrefixForIpv6SourceNat=data.get("EnablePrefixForIpv6SourceNat"),
            IpamPools=data.get("IpamPools"),
        )
        _response = self.client.create_load_balancer(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateLoadBalancerOutput(**_response)

        self.sessionize(response.LoadBalancers[0])
        return cast("LoadBalancer", response.LoadBalancers[0])

    def delete(self, LoadBalancerArn: str) -> None:
        """
        Deletes the specified Application Load Balancer, Network Load Balancer, or
        Gateway Load Balancer. Deleting a load balancer also deletes its listeners.

        Args:
            LoadBalancerArn: The Amazon Resource Name (ARN) of the load balancer.
        """
        args: Dict[str, Any] = dict(LoadBalancerArn=self.serialize(LoadBalancerArn))
        self.client.delete_load_balancer(
            **{k: v for k, v in args.items() if v is not None}
        )

    def get(
        self, *, LoadBalancerArn: str | None = None, Name: str | None = None
    ) -> Optional["LoadBalancer"]:
        """
        Describes the specified load balancers or all of your load balancers.

        Keyword Args:
            LoadBalancerArn: The Amazon Resource Names (ARN) of the load balancer.
            Name: The names of the load balancer.
        """
        args: Dict[str, Any] = dict(
            LoadBalancerArns=self.serialize([LoadBalancerArn]),
            Names=self.serialize([Name]),
        )
        _response = self.client.describe_load_balancers(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeLoadBalancersOutput(**_response)

        if response and response.LoadBalancers:
            self.sessionize(response.LoadBalancers[0])
            return response.LoadBalancers[0]
        return None

    def list(
        self,
        *,
        LoadBalancerArns: List[str] | None = None,
        Names: List[str] | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Describes the specified load balancers or all of your load balancers.

        Keyword Args:
            LoadBalancerArns: The Amazon Resource Names (ARN) of the load balancers. You can specify up to 20 load balancers in
                a single call.
            Names: The names of the load balancers.
        """
        paginator = self.client.get_paginator("describe_load_balancers")
        args: Dict[str, Any] = dict(
            LoadBalancerArns=self.serialize(LoadBalancerArns),
            Names=self.serialize(Names),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DescribeLoadBalancersOutput(**_response)
            if response.LoadBalancers:
                results.extend(response.LoadBalancers)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results

    @load_balancer_attributes_to_dict
    def attributes(self, LoadBalancerArn: str) -> List["LoadBalancerAttribute"]:
        """
        Describes the attributes for the specified Application Load Balancer, Network
        Load Balancer, or Gateway Load Balancer.

        Args:
            LoadBalancerArn: The Amazon Resource Name (ARN) of the load balancer.
        """
        args: Dict[str, Any] = dict(LoadBalancerArn=self.serialize(LoadBalancerArn))
        _response = self.client.describe_load_balancer_attributes(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeLoadBalancerAttributesOutput(**_response)

        results: List["LoadBalancerAttribute"] = None
        if response is not None:
            results = response.Attributes

        self.sessionize(results)
        return cast("List[LoadBalancerAttribute]", results)


class ListenerManager(Boto3ModelManager):
    service_name: str = "elbv2"

    def create(self, model: "Listener", Tags: List[Tag] | None = None) -> "Listener":
        """
        Creates a listener for the specified Application Load Balancer, Network Load
        Balancer, or Gateway Load Balancer.

        Args:
            model: The :py:class:`Listener` to create.

        Keyword Args:
            Tags: The tags to assign to the listener.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            LoadBalancerArn=data.get("LoadBalancerArn"),
            DefaultActions=data.get("DefaultActions"),
            Protocol=data.get("Protocol"),
            Port=data.get("Port"),
            SslPolicy=data.get("SslPolicy"),
            Certificates=data.get("Certificates"),
            AlpnPolicy=data.get("AlpnPolicy"),
            Tags=self.serialize(Tags),
            MutualAuthentication=data.get("MutualAuthentication"),
        )
        _response = self.client.create_listener(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateListenerOutput(**_response)

        self.sessionize(response.Listeners[0])
        return cast("Listener", response.Listeners[0])

    def update(self, model: "Listener") -> "Listener":
        """
        Replaces the specified properties of the specified listener. Any properties that
        you do not specify remain unchanged.

        Args:
            model: The :py:class:`Listener` to update.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            ListenerArn=data.get("ListenerArn"),
            Port=data.get("Port"),
            Protocol=data.get("Protocol"),
            SslPolicy=data.get("SslPolicy"),
            Certificates=data.get("Certificates"),
            DefaultActions=data.get("DefaultActions"),
            AlpnPolicy=data.get("AlpnPolicy"),
            MutualAuthentication=data.get("MutualAuthentication"),
        )
        _response = self.client.modify_listener(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = ModifyListenerOutput(**_response)

        self.sessionize(response.Listeners[0])
        return cast("Listener", response.Listeners[0])

    def delete(self, ListenerArn: str) -> None:
        """
        Deletes the specified listener.

        Args:
            ListenerArn: The Amazon Resource Name (ARN) of the listener.
        """
        args: Dict[str, Any] = dict(ListenerArn=self.serialize(ListenerArn))
        self.client.delete_listener(**{k: v for k, v in args.items() if v is not None})

    def get(self, ListenerArn: str) -> Optional["Listener"]:
        """
        Describes the specified listeners or the listeners for the specified Application
        Load Balancer, Network Load Balancer, or Gateway Load Balancer. You must specify
        either a load balancer or one or more listeners.

        Args:
            ListenerArn: The Amazon Resource Names (ARN) of the listener.
        """
        args: Dict[str, Any] = dict(ListenerArns=self.serialize([ListenerArn]))
        _response = self.client.describe_listeners(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeListenersOutput(**_response)

        if response and response.Listeners:
            self.sessionize(response.Listeners[0])
            return response.Listeners[0]
        return None

    def list(
        self,
        *,
        LoadBalancerArn: str | None = None,
        ListenerArns: List[str] | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Describes the specified listeners or the listeners for the specified Application
        Load Balancer, Network Load Balancer, or Gateway Load Balancer. You must specify
        either a load balancer or one or more listeners.

        Keyword Args:
            LoadBalancerArn: The Amazon Resource Name (ARN) of the load balancer.
            ListenerArns: The Amazon Resource Names (ARN) of the listeners.
        """
        paginator = self.client.get_paginator("describe_listeners")
        args: Dict[str, Any] = dict(
            LoadBalancerArn=self.serialize(LoadBalancerArn),
            ListenerArns=self.serialize(ListenerArns),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DescribeListenersOutput(**_response)
            if response.Listeners:
                results.extend(response.Listeners)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results


class RuleManager(Boto3ModelManager):
    service_name: str = "elbv2"

    def create(
        self, model: "Rule", ListenerArn: str, Tags: List[Tag] | None = None
    ) -> "Rule":
        """
        Creates a rule for the specified listener. The listener must be associated with
        an Application Load Balancer.

        Args:
            model: The :py:class:`Rule` to create.
            ListenerArn: The Amazon Resource Name (ARN) of the listener.

        Keyword Args:
            Tags: The tags to assign to the rule.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            ListenerArn=self.serialize(ListenerArn),
            Conditions=data.get("Conditions"),
            Priority=data.get("Priority"),
            Actions=data.get("Actions"),
            Tags=self.serialize(Tags),
        )
        _response = self.client.create_rule(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateRuleOutput(**_response)

        self.sessionize(response.Rules[0])
        return cast("Rule", response.Rules[0])

    def update(self, model: "Rule") -> "Rule":
        """
        Replaces the specified properties of the specified rule. Any properties that you
        do not specify are unchanged.

        Args:
            model: The :py:class:`Rule` to update.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            RuleArn=data.get("RuleArn"),
            Conditions=data.get("Conditions"),
            Actions=data.get("Actions"),
        )
        _response = self.client.modify_rule(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = ModifyRuleOutput(**_response)

        self.sessionize(response.Rules[0])
        return cast("Rule", response.Rules[0])

    def delete(self, RuleArn: str) -> None:
        """
        Deletes the specified rule.

        Args:
            RuleArn: The Amazon Resource Name (ARN) of the rule.
        """
        args: Dict[str, Any] = dict(RuleArn=self.serialize(RuleArn))
        self.client.delete_rule(**{k: v for k, v in args.items() if v is not None})

    def get(self, RuleArn: str) -> Optional["Rule"]:
        """
        Describes the specified rules or the rules for the specified listener. You must
        specify either a listener or one or more rules.

        Args:
            RuleArn: The Amazon Resource Names (ARN) of the rule.
        """
        args: Dict[str, Any] = dict(RuleArns=self.serialize([RuleArn]))
        _response = self.client.describe_rules(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeRulesOutput(**_response)

        if response and response.Rules:
            self.sessionize(response.Rules[0])
            return response.Rules[0]
        return None

    def list(
        self, *, ListenerArn: str | None = None, RuleArns: List[str] | None = None
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Describes the specified rules or the rules for the specified listener. You must
        specify either a listener or one or more rules.

        Keyword Args:
            ListenerArn: The Amazon Resource Name (ARN) of the listener.
            RuleArns: The Amazon Resource Names (ARN) of the rules.
        """
        paginator = self.client.get_paginator("describe_rules")
        args: Dict[str, Any] = dict(
            ListenerArn=self.serialize(ListenerArn), RuleArns=self.serialize(RuleArns)
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DescribeRulesOutput(**_response)
            if response.Rules:
                results.extend(response.Rules)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results


class TargetGroupManager(Boto3ModelManager):
    service_name: str = "elbv2"

    def create(
        self, model: "TargetGroup", Name: str, Tags: List[Tag] | None = None
    ) -> "TargetGroup":
        """
        Creates a target group.

        Args:
            model: The :py:class:`TargetGroup` to create.
            Name: The name of the target group.

        Keyword Args:
            Tags: The tags to assign to the target group.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Name=self.serialize(Name),
            Protocol=data.get("Protocol"),
            ProtocolVersion=data.get("ProtocolVersion"),
            Port=data.get("Port"),
            VpcId=data.get("VpcId"),
            HealthCheckProtocol=data.get("HealthCheckProtocol"),
            HealthCheckPort=data.get("HealthCheckPort"),
            HealthCheckEnabled=data.get("HealthCheckEnabled"),
            HealthCheckPath=data.get("HealthCheckPath"),
            HealthCheckIntervalSeconds=data.get("HealthCheckIntervalSeconds"),
            HealthCheckTimeoutSeconds=data.get("HealthCheckTimeoutSeconds"),
            HealthyThresholdCount=data.get("HealthyThresholdCount"),
            UnhealthyThresholdCount=data.get("UnhealthyThresholdCount"),
            Matcher=data.get("Matcher"),
            TargetType=data.get("TargetType"),
            Tags=self.serialize(Tags),
            IpAddressType=data.get("IpAddressType"),
        )
        _response = self.client.create_target_group(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateTargetGroupOutput(**_response)

        self.sessionize(response.TargetGroups[0])
        return cast("TargetGroup", response.TargetGroups[0])

    def update(self, model: "TargetGroup") -> "TargetGroup":
        """
        Modifies the health checks used when evaluating the health state of the targets
        in the specified target group.

        Args:
            model: The :py:class:`TargetGroup` to update.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            TargetGroupArn=data.get("TargetGroupArn"),
            HealthCheckProtocol=data.get("HealthCheckProtocol"),
            HealthCheckPort=data.get("HealthCheckPort"),
            HealthCheckPath=data.get("HealthCheckPath"),
            HealthCheckEnabled=data.get("HealthCheckEnabled"),
            HealthCheckIntervalSeconds=data.get("HealthCheckIntervalSeconds"),
            HealthCheckTimeoutSeconds=data.get("HealthCheckTimeoutSeconds"),
            HealthyThresholdCount=data.get("HealthyThresholdCount"),
            UnhealthyThresholdCount=data.get("UnhealthyThresholdCount"),
            Matcher=data.get("Matcher"),
        )
        _response = self.client.modify_target_group(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = ModifyTargetGroupOutput(**_response)

        self.sessionize(response.TargetGroups[0])
        return cast("TargetGroup", response.TargetGroups[0])

    def delete(self, TargetGroupArn: str) -> None:
        """
        Deletes the specified target group.

        Args:
            TargetGroupArn: The Amazon Resource Name (ARN) of the target group.
        """
        args: Dict[str, Any] = dict(TargetGroupArn=self.serialize(TargetGroupArn))
        self.client.delete_target_group(
            **{k: v for k, v in args.items() if v is not None}
        )

    def get(
        self, *, TargetGroupArn: str | None = None, Name: str | None = None
    ) -> Optional["TargetGroup"]:
        """
        Describes the specified target groups or all of your target groups. By default,
        all target groups are described. Alternatively, you can specify one of the
        following to filter the results: the ARN of the load balancer, the names of one
        or more target groups, or the ARNs of one or more target groups.

        Keyword Args:
            TargetGroupArn: The Amazon Resource Names (ARN) of the target group.
            Name: The name of the target group.
        """
        args: Dict[str, Any] = dict(
            TargetGroupArns=self.serialize([TargetGroupArn]),
            Names=self.serialize([Name]),
        )
        _response = self.client.describe_target_groups(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeTargetGroupsOutput(**_response)

        if response and response.TargetGroups:
            self.sessionize(response.TargetGroups[0])
            return response.TargetGroups[0]
        return None

    def list(
        self,
        *,
        LoadBalancerArn: str | None = None,
        TargetGroupArns: List[str] | None = None,
        Names: List[str] | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Describes the specified target groups or all of your target groups. By default,
        all target groups are described. Alternatively, you can specify one of the
        following to filter the results: the ARN of the load balancer, the names of one
        or more target groups, or the ARNs of one or more target groups.

        Keyword Args:
            LoadBalancerArn: The Amazon Resource Name (ARN) of the load balancer.
            TargetGroupArns: The Amazon Resource Names (ARN) of the target groups.
            Names: The names of the target groups.
        """
        paginator = self.client.get_paginator("describe_target_groups")
        args: Dict[str, Any] = dict(
            LoadBalancerArn=self.serialize(LoadBalancerArn),
            TargetGroupArns=self.serialize(TargetGroupArns),
            Names=self.serialize(Names),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = DescribeTargetGroupsOutput(**_response)
            if response.TargetGroups:
                results.extend(response.TargetGroups)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results

    def targets(
        self,
        TargetGroupArn: str,
        *,
        Targets: List["TargetDescription"] | None = None,
        Include: List[Literal["AnomalyDetection", "All"]] | None = None,
    ) -> List["TargetHealthDescription"] | None:
        """
        Describes the health of the specified targets or all of your targets.

        Args:
            TargetGroupArn: The Amazon Resource Name (ARN) of the target group.

        Keyword Args:
            Targets: The targets.
            Include: Used to include anomaly detection information.
        """
        args: Dict[str, Any] = dict(
            TargetGroupArn=self.serialize(TargetGroupArn),
            Targets=self.serialize(Targets),
            Include=self.serialize(Include),
        )
        _response = self.client.describe_target_health(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeTargetHealthOutput(**_response)

        results: List["TargetHealthDescription"] | None = None
        if response is not None:
            results = response.TargetHealthDescriptions

        self.sessionize(results)
        return cast("List[TargetHealthDescription] | None", results)


# ==============
# Service Models
# ==============


class LoadBalancerState(Boto3Model):
    """
    Information about the state of the load balancer.
    """

    Code: Literal["active", "provisioning", "active_impaired", "failed"] | None = (
        None
    )
    """
    The state code.

    The initial state of the load balancer is ``provisioning``. After the load balancer is fully set up and
    ready to route traffic, its state is ``active``. If load balancer is routing traffic but does not have the resources it
    needs to scale, its state is``active_impaired``. If the load balancer could not be set up, its state is ``failed``.
    """
    Reason: str | None = None
    """
    A description of the state.
    """


class LoadBalancerAddress(Boto3Model):
    """
    Information about a static IP address for a load balancer.
    """

    IpAddress: str | None = None
    """
    The static IP address.
    """
    AllocationId: str | None = None
    """
    [Network Load Balancers] The allocation ID of the Elastic IP address for an
    internal-facing load balancer.
    """
    PrivateIPv4Address: str | None = None
    """
    [Network Load Balancers] The private IPv4 address for an internal load balancer.
    """
    IPv6Address: str | None = None
    """
    [Network Load Balancers] The IPv6 address.
    """


class AvailabilityZone(Boto3Model):
    """
    Information about an Availability Zone.
    """

    ZoneName: str | None = None
    """
    The name of the Availability Zone.
    """
    SubnetId: str | None = None
    """
    The ID of the subnet.

    You can specify one subnet per Availability Zone.
    """
    OutpostId: str | None = None
    """
    [Application Load Balancers on Outposts] The ID of the Outpost.
    """
    LoadBalancerAddresses: List["LoadBalancerAddress"] | None = None
    """
    [Network Load Balancers] If you need static IP addresses for your load balancer, you
    can specify one Elastic IP address per Availability Zone when you create an
    internal-facing load balancer.

    For internal load balancers, you can specify a private IP address from the IPv4
    range of the subnet.
    """
    SourceNatIpv6Prefixes: List[str] | None = None
    """
    [Network Load Balancers with UDP listeners] The IPv6 prefixes to use for source NAT.

    For each subnet, specify an IPv6
    prefix (/80 netmask) from the subnet CIDR block or ``auto_assigned`` to use an IPv6 prefix selected at random from the
    subnet CIDR block.
    """


class ElbV2IpamPools(Boto3Model):
    """
    An IPAM pool is a collection of IP address CIDRs.

    IPAM pools enable you to organize your IP addresses according to your routing and
    security needs.
    """

    Ipv4IpamPoolId: str | None = None
    """
    The ID of the IPv4 IPAM pool.
    """


class LoadBalancer(PrimaryBoto3Model):
    """
    Information about a load balancer.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = LoadBalancerManager

    LoadBalancerName: str
    """
    The name of the load balancer.
    """
    Scheme: Literal["internet-facing", "internal"] | None = "internet-facing"
    """
    The nodes of an Internet-facing load balancer have public IP addresses.

    The DNS name of an Internet-facing load balancer is publicly resolvable to the
    public IP addresses of the nodes. Therefore, Internet-facing load balancers can
    route requests from clients over the internet.
    """
    Type: Literal["application", "network", "gateway"] | None = "application"
    """
    The type of load balancer.
    """
    IpAddressType: Literal["ipv4", "dualstack", "dualstack-without-public-ipv4"] | None = "ipv4"
    """
    The type of IP addresses used for public or private connections by the subnets
    attached to your load balancer.
    """
    LoadBalancerArn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) of the load balancer.
    """
    DNSName: str = Field(default=None, frozen=True)
    """
    The public DNS name of the load balancer.
    """
    CanonicalHostedZoneId: str = Field(default=None, frozen=True)
    """
    The ID of the Amazon Route 53 hosted zone associated with the load balancer.
    """
    CreatedTime: datetime = Field(default=None, frozen=True)
    """
    The date and time the load balancer was created.
    """
    VpcId: str = Field(default=None, frozen=True)
    """
    The ID of the VPC for the load balancer.
    """
    State: LoadBalancerState = Field(default=None, frozen=True)
    """
    The state of the load balancer.
    """
    AvailabilityZones: List["AvailabilityZone"] = Field(
        default_factory=list, frozen=True
    )
    """
    The subnets for the load balancer.
    """
    SecurityGroups: List[str] | None = None
    """
    The IDs of the security groups for the load balancer.
    """
    CustomerOwnedIpv4Pool: str | None = None
    """
    [Application Load Balancers on Outposts] The ID of the customer-owned address pool.
    """
    EnforceSecurityGroupInboundRulesOnPrivateLinkTraffic: str = Field(
        default=None, frozen=True
    )
    """
    Indicates whether to evaluate inbound security group rules for traffic sent to a
    Network Load Balancer through Amazon Web Services PrivateLink.
    """
    EnablePrefixForIpv6SourceNat: Literal["on", "off"] | None = None
    """
    [Network Load Balancers with UDP listeners] Indicates whether to use an IPv6 prefix
    from each subnet for source NAT.

    The
    IP address type must be ``dualstack``. The default value is ``off``.
    """
    IpamPools: ElbV2IpamPools | None = None
    """
    [Application Load Balancers] The IPAM pool in use by the load balancer, if
    configured.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`LoadBalancerArn` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.LoadBalancerArn

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`LoadBalancerArn` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.LoadBalancerArn

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`LoadBalancerName` attribute.

        Returns:
            The name of the model instance.
        """
        return self.LoadBalancerName

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`LoadBalancerArn` attribute.
        """
        return hash(self.LoadBalancerArn)

    @property
    def logging_enabled(self) -> bool:
        """
        Return whether logging is enabled for the load balancer.
        """

        return self.attributes().get("access_logs.s3.enabled") == "true"

    @property
    def logging_bucket(self) -> str:
        """
        Return the name of the S3 bucket that logs are stored in, if any.
        """

        return self.attributes().get("access_logs.s3.bucket")

    @property
    def logging_prefix(self) -> str:
        """
        Return the prefix in the S3 bucket that logs are stored in, if any.
        """

        return self.attributes().get("access_logs.s3.prefix")

    @cached_property
    def listeners(self) -> List["Listener"] | None:
        """
        Return the ARNs of :py:class:`Listener` objects that belong to this load
        balancer, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "load_balancer": self.LoadBalancerArn,
                }
            )
        except AttributeError:
            return []
        return Listener.objects.using(self.session).list(**pk)  # type: ignore[arg-type]

    @cached_property
    def vpc(self) -> Optional["Vpc"]:
        """
        Return the :py:class:`Vpc` object that this load balancer belongs to, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "VpcId": self.VpcId,
                }
            )
        except AttributeError:
            return None
        return Vpc.objects.using(self.session).get(**pk)  # type: ignore[arg-type]

    @cached_property
    def security_groups(self) -> List["SecurityGroup"] | None:
        """
        Return the ARNs of :py:class:`SecurityGroup` objects that belong to this load
        balancer, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "GroupIds": self.SecurityGroups,
                }
            )
        except AttributeError:
            return []
        return SecurityGroup.objects.using(self.session).list(**pk)  # type: ignore[arg-type]

    def attributes(self) -> List["LoadBalancerAttribute"]:
        """
        Return the attributes of the load balancer.

        These are stored separately from the load balancer object
        """

        return (
            cast("LoadBalancerManager", self.objects)  # type: ignore[attr-defined]
            .using(self.session)
            .attributes(
                cast("str", self.arn),
            )
        )


class Certificate(ReadonlyBoto3Model):
    """
    Information about an SSL server certificate.
    """

    CertificateArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the certificate.
    """
    IsDefault: bool | None = None
    """
    Indicates whether the certificate is the default certificate.

    Do not set this value when specifying a certificate as an input. This value is not
    included in the output when describing a listener, but is included when describing
    listener certificates.
    """


class AuthenticateOidcActionConfig(Boto3Model):
    """
    Request parameters when using an identity provider (IdP) that is compliant with
    OpenID Connect (OIDC) to authenticate users.
    """

    Issuer: str
    """
    The OIDC issuer identifier of the IdP.

    This must be a full URL, including the HTTPS protocol, the domain, and the path.
    """
    AuthorizationEndpoint: str
    """
    The authorization endpoint of the IdP.

    This must be a full URL, including the HTTPS protocol, the domain, and the path.
    """
    TokenEndpoint: str
    """
    The token endpoint of the IdP.

    This must be a full URL, including the HTTPS protocol, the domain, and the path.
    """
    UserInfoEndpoint: str
    """
    The user info endpoint of the IdP.

    This must be a full URL, including the HTTPS protocol, the domain, and the path.
    """
    ClientId: str
    """
    The OAuth 2.0 client identifier.
    """
    ClientSecret: str | None = None
    """
    The OAuth 2.0 client secret.

    This parameter is required if you are creating a rule. If you are modifying a rule, you can
    omit this parameter if you set ``UseExistingClientSecret`` to true.
    """
    SessionCookieName: str | None = None
    """
    The name of the cookie used to maintain session information.

    The default is AWSELBAuthSessionCookie.
    """
    Scope: str | None = None
    """
    The set of user claims to be requested from the IdP.

    The default is ``openid``.
    """
    SessionTimeout: int | None = None
    """
    The maximum duration of the authentication session, in seconds.

    The default is 604800 seconds (7 days).
    """
    AuthenticationRequestExtraParams: Dict[str, str] | None = None
    """
    The query parameters (up to 10) to include in the redirect request to the
    authorization endpoint.
    """
    OnUnauthenticatedRequest: Literal["deny", "allow", "authenticate"] | None = None
    """
    The behavior if the user is not authenticated.

    The following are possible values:
    """
    UseExistingClientSecret: bool | None = None
    """
    Indicates whether to use the existing client secret when modifying a rule.

    If you are creating a rule, you can omit this parameter or set it to false.
    """


class AuthenticateCognitoActionConfig(Boto3Model):
    """
    Request parameters to use when integrating with Amazon Cognito to authenticate
    users.
    """

    UserPoolArn: str
    """
    The Amazon Resource Name (ARN) of the Amazon Cognito user pool.
    """
    UserPoolClientId: str
    """
    The ID of the Amazon Cognito user pool client.
    """
    UserPoolDomain: str
    """
    The domain prefix or fully-qualified domain name of the Amazon Cognito user pool.
    """
    SessionCookieName: str | None = None
    """
    The name of the cookie used to maintain session information.

    The default is AWSELBAuthSessionCookie.
    """
    Scope: str | None = None
    """
    The set of user claims to be requested from the IdP.

    The default is ``openid``.
    """
    SessionTimeout: int | None = None
    """
    The maximum duration of the authentication session, in seconds.

    The default is 604800 seconds (7 days).
    """
    AuthenticationRequestExtraParams: Dict[str, str] | None = None
    """
    The query parameters (up to 10) to include in the redirect request to the
    authorization endpoint.
    """
    OnUnauthenticatedRequest: Literal["deny", "allow", "authenticate"] | None = None
    """
    The behavior if the user is not authenticated.

    The following are possible values:
    """


class RedirectActionConfig(Boto3Model):
    """
    Information about a redirect action.

    A URI consists of the following components: protocol://hostname:port/path?query. You must modify at least one of the
    following components to avoid a redirect loop: protocol, hostname, port, or path. Any components that you do not modify
    retain their original values.

    You can reuse URI components using the following reserved keywords:

    * #{protocol}
    * #{host}
    * #{port}
    * #{path} (the leading "/" is removed)
    * #{query}

    For example, you can change the path to "/new/#{path}", the hostname to "example.#{host}", or the query to
    "#{query}&value=xyz".
    """

    Protocol: str | None = None
    """
    The protocol.

    You can specify HTTP, HTTPS, or #{protocol}. You can redirect HTTP to HTTP, HTTP to
    HTTPS, and HTTPS to HTTPS. You can't redirect HTTPS to HTTP.
    """
    Port: str | None = None
    """
    The port.

    You can specify a value from 1 to 65535 or #{port}.
    """
    Host: str | None = None
    """
    The hostname.

    This component is not percent-encoded. The hostname can contain #{host}.
    """
    Path: str | None = None
    """
    The absolute path, starting with the leading "/".

    This component is not percent-encoded. The path can contain #{host}, #{path}, and
    #{port}.
    """
    Query: str | None = None
    """
    The query parameters, URL-encoded when necessary, but not percent-encoded.

    Do not include the leading "?", as it is automatically added. You can specify any of
    the reserved keywords.
    """
    StatusCode: Literal["HTTP_301", "HTTP_302"]
    """
    The HTTP redirect code.

    The redirect is either permanent (HTTP 301) or temporary (HTTP 302).
    """


class FixedResponseActionConfig(Boto3Model):
    """
    Information about an action that returns a custom HTTP response.
    """

    MessageBody: str | None = None
    """
    The message.
    """
    StatusCode: str
    """
    The HTTP response code (2XX, 4XX, or 5XX).
    """
    ContentType: str | None = None
    """
    The content type.
    """


class TargetGroupTuple(Boto3Model):
    """
    Information about how traffic will be distributed between multiple target groups in
    a forward rule.
    """

    TargetGroupArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the target group.
    """
    Weight: int | None = None
    """
    The weight.

    The range is 0 to 999.
    """


class ElbV2TargetGroupStickinessConfig(Boto3Model):
    """
    Information about the target group stickiness for a rule.
    """

    Enabled: bool | None = None
    """
    Indicates whether target group stickiness is enabled.
    """
    DurationSeconds: int | None = None
    """
    The time period, in seconds, during which requests from a client should be routed to
    the same target group.

    The range is 1-604800 seconds (7 days). You must specify this value when enabling
    target group stickiness.
    """


class ForwardActionConfig(Boto3Model):
    """
    Information about a forward action.
    """

    TargetGroups: List["TargetGroupTuple"] | None = None
    """
    The target groups.

    For Network Load Balancers, you can specify a single target group.
    """
    TargetGroupStickinessConfig: ElbV2TargetGroupStickinessConfig | None = None
    """
    The target group stickiness for the rule.
    """


class Action(Boto3Model):
    """
    Information about an action.

    Each rule must include exactly one of the following types of actions: ``forward``, ``fixed-response``, or ``redirect``,
    and it must be the last action to be performed.
    """

    ActionType: Literal[
        "forward",
        "authenticate-oidc",
        "authenticate-cognito",
        "redirect",
        "fixed-response",
    ] = Field(default=None, alias="Type")
    """
    The type of action.
    """
    TargetGroupArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the target group.

    Specify only when ``Type`` is ``forward`` and you want to route to a
    single target group. To route to one or more target groups, use ``ForwardConfig`` instead.
    """
    AuthenticateOidcConfig: AuthenticateOidcActionConfig | None = None
    """
    [HTTPS listeners] Information about an identity provider that is compliant with
    OpenID Connect (OIDC).

    Specify only when
    ``Type`` is ``authenticate-oidc``.
    """
    AuthenticateCognitoConfig: AuthenticateCognitoActionConfig | None = None
    """
    [HTTPS listeners] Information for using Amazon Cognito to authenticate users.

    Specify only when ``Type`` is
    ``authenticate-cognito``.
    """
    Order: int | None = None
    """
    The order for the action.

    This value is required for rules with multiple actions. The action with the lowest
    value for order is performed first.
    """
    RedirectConfig: RedirectActionConfig | None = None
    """
    [Application Load Balancer] Information for creating a redirect action.

    Specify only when ``Type`` is ``redirect``.
    """
    FixedResponseConfig: FixedResponseActionConfig | None = None
    """
    [Application Load Balancer] Information for creating an action that returns a custom
    HTTP response.

    Specify only when
    ``Type`` is ``fixed-response``.
    """
    ForwardConfig: ForwardActionConfig | None = None
    """
    Information for creating an action that distributes requests among one or more
    target groups.

    For Network Load
    Balancers, you can specify a single target group. Specify only when ``Type`` is ``forward``. If you specify both
    ``ForwardConfig`` and ``TargetGroupArn``, you can specify only one target group using ``ForwardConfig`` and it must be
    the same target group specified in ``TargetGroupArn``.
    """


class MutualAuthenticationAttributes(Boto3Model):
    """
    Information about the mutual authentication attributes of a listener.
    """

    Mode: str | None = None
    """
    The client certificate handling method.

    Options are ``off``, ``passthrough`` or ``verify``. The default value is
    ``off``.
    """
    TrustStoreArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the trust store.
    """
    IgnoreClientCertificateExpiry: bool | None = None
    """
    Indicates whether expired client certificates are ignored.
    """
    TrustStoreAssociationStatus: Literal["active", "removed"] | None = None
    """
    Indicates a shared trust stores association status.
    """
    AdvertiseTrustStoreCaNames: Literal["on", "off"] | None = None
    """
    Indicates whether trust store CA certificate names are advertised.
    """


class Listener(PrimaryBoto3Model):
    """
    Information about a listener.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = ListenerManager

    LoadBalancerArn: str
    """
    The Amazon Resource Name (ARN) of the load balancer.
    """
    Port: int
    """
    The port on which the load balancer is listening.
    """
    Protocol: Literal["HTTP", "HTTPS", "TCP", "TLS", "UDP", "TCP_UDP", "GENEVE"] | None = "HTTPS"
    """
    The protocol for connections from clients to the load balancer.
    """
    ListenerArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the listener.
    """
    Certificates: List["Certificate"] | None = None
    """
    [HTTPS or TLS listener] The default certificate for the listener.
    """
    SslPolicy: str | None = None
    """
    [HTTPS or TLS listener] The security policy that defines which protocols and ciphers
    are supported.
    """
    DefaultActions: List["Action"] | None = None
    """
    The default actions for the listener.
    """
    AlpnPolicy: List[str] | None = None
    """
    [TLS listener] The name of the Application-Layer Protocol Negotiation (ALPN) policy.
    """
    MutualAuthentication: MutualAuthenticationAttributes | None = None
    """
    The mutual authentication configuration information.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`ListenerArn` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.ListenerArn

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`ListenerArn`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.ListenerArn

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`ListenerArn` attribute.
        """
        return hash(self.ListenerArn)

    @cached_property
    def load_balancer(self) -> Optional["LoadBalancer"]:
        """
        Return the :py:class:`LoadBalancer` object that this listener belongs to, if
        any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "LoadBalancerArn": self.LoadBalancerArn,
                }
            )
        except AttributeError:
            return None
        return LoadBalancer.objects.using(self.session).get(**pk)  # type: ignore[arg-type]

    @cached_property
    def rules(self) -> List["Rule"] | None:
        """
        Return the ARNs of :py:class:`Rule` objects that belong to this listener, if
        any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "ListenerArn": self.ListenerArn,
                }
            )
        except AttributeError:
            return []
        return Rule.objects.using(self.session).list(**pk)  # type: ignore[arg-type]


class HostHeaderConditionConfig(Boto3Model):
    """
    Information about a host header condition.
    """

    Values: List[str] | None = None
    """
    The host names.

    The maximum size of each name is 128 characters. The comparison is case insensitive. The following
    wildcard characters are supported: * (matches 0 or more characters) and ? (matches exactly 1 character). You must
    include at least one "." character. You can include only alphabetical characters after the final "." character.
    """


class PathPatternConditionConfig(Boto3Model):
    """
    Information about a path pattern condition.
    """

    Values: List[str] | None = None
    """
    The path patterns to compare against the request URL.

    The maximum size of each string is 128 characters. The comparison is case sensitive.
    The following wildcard characters are supported: * (matches 0 or more characters)
    and ? (matches exactly 1 character).
    """


class HttpHeaderConditionConfig(Boto3Model):
    """
    Information about an HTTP header condition.

    There is a set of standard HTTP header fields. You can also define custom HTTP
    header fields.
    """

    HttpHeaderName: str | None = None
    """
    The name of the HTTP header field.

    The maximum size is 40 characters. The header name is case insensitive. The allowed
    characters are specified by RFC 7230. Wildcards are not supported.
    """
    Values: List[str] | None = None
    """
    The strings to compare against the value of the HTTP header.

    The maximum size of each string is 128 characters. The comparison strings are case
    insensitive. The following wildcard characters are supported: * (matches 0 or more
    characters) and ? (matches exactly 1 character).
    """


class QueryStringKeyValuePair(Boto3Model):
    """
    Information about a key/value pair.
    """

    Key: str | None = None
    """
    The key.

    You can omit the key.
    """
    Value: str | None = None
    """
    The value.
    """


class QueryStringConditionConfig(Boto3Model):
    """
    Information about a query string condition.

    The query string component of a URI starts after the first '?' character and is
    terminated by either a '#' character or the end of the URI. A typical query string
    contains key/value pairs separated by '&' characters. The allowed characters are
    specified by RFC 3986. Any character can be percentage encoded.
    """

    Values: List["QueryStringKeyValuePair"] | None = None
    """
    The key/value pairs or values to find in the query string.

    The maximum size of each string is 128 characters. The
    comparison is case insensitive. The following wildcard characters are supported: * (matches 0 or more characters) and ?
    (matches exactly 1 character). To search for a literal '*' or '?' character in a query string, you must escape these
    characters in ``Values`` using a '' character.
    """


class HttpRequestMethodConditionConfig(Boto3Model):
    """
    Information about an HTTP method condition.

    HTTP defines a set of request methods, also referred to as HTTP verbs. For more
    information, see the
    `HTTP Method Registry <https://www.iana.org/assignments/http-methods/http-methods.xhtml>`_.
    You can also define custom HTTP methods.
    """

    Values: List[str] | None = None
    """
    The name of the request method.

    The maximum size is 40 characters. The allowed characters are A-Z, hyphen (-), and
    underscore (_). The comparison is case sensitive. Wildcards are not supported;
    therefore, the method name must be an exact match.
    """


class SourceIpConditionConfig(Boto3Model):
    """
    Information about a source IP condition.

    You can use this condition to route based on the IP address of the source that
    connects to the load balancer. If a client is behind a proxy, this is the IP address
    of the proxy not the IP address of the client.
    """

    Values: List[str] | None = None
    """
    The source IP addresses, in CIDR format.

    You can use both IPv4 and IPv6 addresses. Wildcards are not supported.
    """


class RuleCondition(Boto3Model):
    """
    Information about a condition for a rule.

    Each rule can optionally include up to one of each of the following conditions: ``http-request-method``, ``host-
    header``, ``path-pattern``, and ``source-ip``. Each rule can also optionally include one or more of each of the
    following conditions: ``http-header`` and ``query-string``. Note that the value for a condition can't be empty.

    For more information, see `Quotas for your Application Load
    Balancers <https://docs.aws.amazon.com/elasticloadbalancing/latest/application/load-balancer-limits.html>`_.
    """

    FieldName: str = Field(default=None, alias="Field")
    """
    The field in the HTTP request.

    The following are the possible values:
    """
    Values: List[str] | None = None
    """
    The condition value.

    Specify only when ``Field`` is ``host-header`` or ``path-pattern``. Alternatively, to specify
    multiple host names or multiple path patterns, use ``HostHeaderConfig`` or ``PathPatternConfig``.
    """
    HostHeaderConfig: HostHeaderConditionConfig | None = None
    """
    Information for a host header condition.

    Specify only when ``Field`` is ``host-header``.
    """
    PathPatternConfig: PathPatternConditionConfig | None = None
    """
    Information for a path pattern condition.

    Specify only when ``Field`` is ``path-pattern``.
    """
    HttpHeaderConfig: HttpHeaderConditionConfig | None = None
    """
    Information for an HTTP header condition.

    Specify only when ``Field`` is ``http-header``.
    """
    QueryStringConfig: QueryStringConditionConfig | None = None
    """
    Information for a query string condition.

    Specify only when ``Field`` is ``query-string``.
    """
    HttpRequestMethodConfig: HttpRequestMethodConditionConfig | None = None
    """
    Information for an HTTP method condition.

    Specify only when ``Field`` is ``http-request-method``.
    """
    SourceIpConfig: SourceIpConditionConfig | None = None
    """
    Information for a source IP condition.

    Specify only when ``Field`` is ``source-ip``.
    """


class Rule(PrimaryBoto3Model):
    """
    Information about a rule.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = RuleManager

    Priority: str | None = "1"
    """
    The priority.
    """
    Conditions: List["RuleCondition"]
    """
    The conditions.

    Each rule can include zero or one of the following conditions: ``http-request-method``, ``host-header``,
    ``path-pattern``, and ``source-ip``, and zero or more of the following conditions: ``http-header`` and ``query-string``.
    """
    Actions: List["Action"]
    """
    The actions.

    Each rule must include exactly one of the following types of actions: ``forward``, ``redirect``, or
    ``fixed-response``, and it must be the last action to be performed.
    """
    IsDefault: bool | None = False
    """
    Indicates whether this is the default rule.
    """
    RuleArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the rule.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`RuleArn` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.RuleArn

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`RuleArn`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.RuleArn

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`RuleArn` attribute.
        """
        return hash(self.RuleArn)


class ResponseCodeMatcher(Boto3Model):
    """
    The codes to use when checking for a successful response from a target.

    If the protocol version is gRPC, these are gRPC codes. Otherwise, these are HTTP
    codes.
    """

    HttpCode: str | None = None
    """
    For Application Load Balancers, you can specify values between 200 and 499, with the
    default value being 200.

    You can specify multiple values (for example, "200,202") or a range of values (for
    example, "200-299").
    """
    GrpcCode: str | None = None
    """
    You can specify values between 0 and 99.

    You can specify multiple values (for example, "0,1") or a range of values (for
    example, "0-5"). The default value is 12.
    """


class TargetDescription(Boto3Model):
    """
    Information about a target.
    """

    Id: str
    """
    The ID of the target.

    If the target type of the target group is ``instance``, specify an instance ID. If the target type
    is ``ip``, specify an IP address. If the target type is ``lambda``, specify the ARN of the Lambda function. If the
    target type is ``alb``, specify the ARN of the Application Load Balancer target.
    """
    Port: int | None = None
    """
    The port on which the target is listening.

    If the target group protocol is GENEVE, the supported port is 6081. If the
    target type is ``alb``, the targeted Application Load Balancer must have at least one listener whose port matches the
    target group port. This parameter is not used if the target is a Lambda function.
    """
    AvailabilityZone: str | None = None
    """
    An Availability Zone or ``all``.

    This determines whether the target receives traffic from the load balancer nodes in
    the specified Availability Zone or from all enabled Availability Zones for the load
    balancer.
    """


class TargetGroup(PrimaryBoto3Model):
    """
    Information about a target group.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = TargetGroupManager

    TargetGroupName: str
    """
    The name of the target group.
    """
    Protocol: Literal["HTTP", "HTTPS", "TCP", "TLS", "UDP", "TCP_UDP", "GENEVE"] | None = "HTTPS"
    """
    The protocol to use for routing traffic to the targets.
    """
    Port: int
    """
    The port on which the targets are listening.

    This parameter is not used if the target is a Lambda function.
    """
    VpcId: str
    """
    The ID of the VPC for the targets.
    """
    TargetType: Literal["instance", "ip", "lambda", "alb"] | None = "ip"
    """
    The type of target that you must specify when registering targets with this target
    group.

    The possible values are
    ``instance`` (register targets by instance ID), ``ip`` (register targets by IP address), ``lambda`` (register a single
    Lambda function as a target), or ``alb`` (register a single Application Load Balancer as a target).
    """
    IpAddressType: Literal["ipv4", "ipv6"] | None = "ipv4"
    """
    The IP address type.

    The default value is ``ipv4``.
    """
    TargetGroupArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the target group.
    """
    HealthCheckProtocol: Literal["HTTP", "HTTPS", "TCP", "TLS", "UDP", "TCP_UDP", "GENEVE"] | None = None
    """
    The protocol to use to connect with the target.

    The GENEVE, TLS, UDP, and TCP_UDP protocols are not supported for health checks.
    """
    HealthCheckPort: str | None = None
    """
    The port to use to connect with the target.
    """
    HealthCheckEnabled: bool | None = None
    """
    Indicates whether health checks are enabled.
    """
    HealthCheckIntervalSeconds: int | None = None
    """
    The approximate amount of time, in seconds, between health checks of an individual
    target.
    """
    HealthCheckTimeoutSeconds: int | None = None
    """
    The amount of time, in seconds, during which no response means a failed health
    check.
    """
    HealthyThresholdCount: int | None = None
    """
    The number of consecutive health checks successes required before considering an
    unhealthy target healthy.
    """
    UnhealthyThresholdCount: int | None = None
    """
    The number of consecutive health check failures required before considering the
    target unhealthy.
    """
    HealthCheckPath: str | None = None
    """
    The destination for health checks on the targets.
    """
    Matcher: ResponseCodeMatcher | None = None
    """
    The HTTP or gRPC codes to use when checking for a successful response from a target.
    """
    LoadBalancerArns: List[str] | None = None
    """
    The Amazon Resource Name (ARN) of the load balancer that routes traffic to this
    target group.

    You can use each target group with only one load balancer.
    """
    ProtocolVersion: str | None = None
    """
    [HTTP/HTTPS protocol] The protocol version.

    The possible values are ``GRPC``, ``HTTP1``, and ``HTTP2``.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`TargetGroupArn` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.TargetGroupArn

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`TargetGroupArn` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.TargetGroupArn

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`TargetGroupName` attribute.

        Returns:
            The name of the model instance.
        """
        return self.TargetGroupName

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`TargetGroupArn` attribute.
        """
        return hash(self.TargetGroupArn)

    @cached_property
    def load_balancers(self) -> List["LoadBalancer"] | None:
        """
        Return the ARNs of :py:class:`LoadBalancer` objects that this target group
        belongs to, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "LoadBalancerArns": self.LoadBalancerArns,
                }
            )
        except AttributeError:
            return []
        return LoadBalancer.objects.using(self.session).list(**pk)  # type: ignore[arg-type]

    @cached_property
    def vpc(self) -> Optional["Vpc"]:
        """
        Return the :py:class:`Vpc` object that this target group belongs to, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "VpcId": self.VpcId,
                }
            )
        except AttributeError:
            return None
        return Vpc.objects.using(self.session).get(**pk)  # type: ignore[arg-type]

    def targets(self) -> List["TargetHealthDescription"] | None:
        """
        Return the targets for the target group.
        """

        return (
            cast("TargetGroupManager", self.objects)  # type: ignore[attr-defined]
            .using(self.session)
            .targets(
                cast("str", self.arn),
            )
        )


# =======================
# Request/Response Models
# =======================


class SubnetMapping(Boto3Model):
    """
    Information about a subnet mapping.
    """

    SubnetId: str
    """
    The ID of the subnet.
    """
    AllocationId: str | None = None
    """
    [Network Load Balancers] The allocation ID of the Elastic IP address for an
    internet-facing load balancer.
    """
    PrivateIPv4Address: str | None = None
    """
    [Network Load Balancers] The private IPv4 address for an internal load balancer.
    """
    IPv6Address: str | None = None
    """
    [Network Load Balancers] The IPv6 address.
    """
    SourceNatIpv6Prefix: str | None = None
    """
    [Network Load Balancers with UDP listeners] The IPv6 prefix to use for source NAT.

    Specify an IPv6 prefix (/80 netmask)
    from the subnet CIDR block or ``auto_assigned`` to use an IPv6 prefix selected at random from the subnet CIDR block.
    """


class CreateLoadBalancerOutput(Boto3Model):
    LoadBalancers: List["LoadBalancer"] | None = None
    """
    Information about the load balancer.
    """


class DeleteLoadBalancerOutput(Boto3Model):
    pass


class DescribeLoadBalancersOutput(Boto3Model):
    LoadBalancers: List["LoadBalancer"] | None = None
    """
    Information about the load balancers.
    """
    NextMarker: str | None = None
    """
    If there are additional results, this is the marker for the next set of results.

    Otherwise, this is null.
    """


class LoadBalancerAttribute(Boto3Model):
    """
    Information about a load balancer attribute.
    """

    Key: str | None = None
    """
    The name of the attribute.
    """
    Value: str | None = None
    """
    The value of the attribute.
    """


class DescribeLoadBalancerAttributesOutput(Boto3Model):
    Attributes: List["LoadBalancerAttribute"] | None = None
    """
    Information about the load balancer attributes.
    """


class CreateListenerOutput(Boto3Model):
    Listeners: List["Listener"] | None = None
    """
    Information about the listener.
    """


class ModifyListenerOutput(Boto3Model):
    Listeners: List["Listener"] | None = None
    """
    Information about the modified listener.
    """


class DeleteListenerOutput(Boto3Model):
    pass


class DescribeListenersOutput(Boto3Model):
    Listeners: List["Listener"] | None = None
    """
    Information about the listeners.
    """
    NextMarker: str | None = None
    """
    If there are additional results, this is the marker for the next set of results.

    Otherwise, this is null.
    """


class CreateRuleOutput(Boto3Model):
    Rules: List["Rule"] | None = None
    """
    Information about the rule.
    """


class ModifyRuleOutput(Boto3Model):
    Rules: List["Rule"] | None = None
    """
    Information about the modified rule.
    """


class DeleteRuleOutput(Boto3Model):
    pass


class DescribeRulesOutput(Boto3Model):
    Rules: List["Rule"] | None = None
    """
    Information about the rules.
    """
    NextMarker: str | None = None
    """
    If there are additional results, this is the marker for the next set of results.

    Otherwise, this is null.
    """


class CreateTargetGroupOutput(Boto3Model):
    TargetGroups: List["TargetGroup"] | None = None
    """
    Information about the target group.
    """


class ModifyTargetGroupOutput(Boto3Model):
    TargetGroups: List["TargetGroup"] | None = None
    """
    Information about the modified target group.
    """


class DeleteTargetGroupOutput(Boto3Model):
    pass


class DescribeTargetGroupsOutput(Boto3Model):
    TargetGroups: List["TargetGroup"] | None = None
    """
    Information about the target groups.
    """
    NextMarker: str | None = None
    """
    If there are additional results, this is the marker for the next set of results.

    Otherwise, this is null.
    """


class TargetHealthInfo(Boto3Model):
    """
    Information about the current health of a target.
    """

    State: Literal["initial", "healthy", "unhealthy", "unhealthy.draining", "unused", "draining", "unavailable"] | None = None
    """
    The state of the target.
    """
    Reason: Literal["Elb.RegistrationInProgress", "Elb.InitialHealthChecking", "Target.ResponseCodeMismatch", "Target.Timeout", "Target.FailedHealthChecks", "Target.NotRegistered", "Target.NotInUse", "Target.DeregistrationInProgress", "Target.InvalidState", "Target.IpUnusable", "Target.HealthCheckDisabled", "Elb.InternalError"] | None = None
    """
    The reason code.
    """
    Description: str | None = None
    """
    A description of the target health that provides additional details.

    If the state is ``healthy``, a description is not
    provided.
    """


class AnomalyDetectionInfo(Boto3Model):
    """
    Information about anomaly detection and mitigation.
    """

    Result: Literal["anomalous", "normal"] | None = None
    """
    The latest anomaly detection result.
    """
    MitigationInEffect: Literal["yes", "no"] | None = None
    """
    Indicates whether anomaly mitigation is in progress.
    """


class ElbV2AdministrativeOverride(Boto3Model):
    """
    Information about the override status applied to a target.
    """

    State: Literal["unknown", "no_override", "zonal_shift_active", "zonal_shift_delegated_to_dns"] | None = None
    """
    The state of the override.
    """
    Reason: Literal["AdministrativeOverride.Unknown", "AdministrativeOverride.NoOverride", "AdministrativeOverride.ZonalShiftActive", "AdministrativeOverride.ZonalShiftDelegatedToDns"] | None = None
    """
    The reason code for the state.
    """
    Description: str | None = None
    """
    A description of the override state that provides additional details.
    """


class TargetHealthDescription(Boto3Model):
    """
    Information about the health of a target.
    """

    Target: TargetDescription | None = None
    """
    The description of the target.
    """
    HealthCheckPort: str | None = None
    """
    The port to use to connect with the target.
    """
    TargetHealth: TargetHealthInfo | None = None
    """
    The health information for the target.
    """
    AnomalyDetection: AnomalyDetectionInfo | None = None
    """
    The anomaly detection result for the target.
    """
    AdministrativeOverride: ElbV2AdministrativeOverride | None = None
    """
    The administrative override information for the target.
    """


class DescribeTargetHealthOutput(Boto3Model):
    TargetHealthDescriptions: List["TargetHealthDescription"] | None = None
    """
    Information about the health of the targets.
    """



import re
import subprocess
import tempfile
import warnings
from collections import OrderedDict
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Final, List, Set, Type, cast

import boto3
import botocore.model
import botocore.session
from docformatter.format import Formatter

from .docstring import DocumentationFormatter, FormatterArgs
from .exceptions import ModelHasNoMembersError, NoPythonTypeError
from .methods import (
    CreateMethodGenerator,
    DeleteMethodGenerator,
    GeneralMethodGenerator,
    GetManyMethodGenerator,
    GetMethodGenerator,
    ListMethodGenerator,
    ManagerMethodGenerator,
    ModelManagerMethodGenerator,
    ModelPropertyGenerator,
    ModelRelationGenerator,
    PartialUpdateMethodGenerator,
    UpdateMethodGenerator,
)
from .models import (
    ManagerDefinition,
    ManagerMethodDefinition,
    ModelAttributeDefinition,
    ModelDefinition,
    ServiceDefinition,
)
from .shapes import PythonTypeShapeConverter
from .sphinx import ServiceSphinxDocBuilder


@dataclass
class ModelTagsDAO:
    """
    A data access object for how to handle tags in a model.

    """

    #: The new set of base classes for this model, possibly including the
    #: ``TagsDictMixin`` class.
    base_class: str
    #: The name of the tag class.  This is the class that will be used to
    #: represent the tags in the model.  This will be ``None`` if there are no
    #: tags in the model.
    tag_class: str | None = None


class AbstractGenerator:
    def __init__(self, service_generator: "ServiceGenerator") -> None:
        session = botocore.session.get_session()
        #: The :py:class:`ServiceGenerator` we're generating models for.
        self.service_generator = service_generator
        #: The name of the AWS service we're generating models for.
        self.service_name = service_generator.aws_service_name
        #: The botocraft service definition for our service.
        self.service_def = service_generator.service_def
        #: The botocraft interface definition.  We collect things we need to
        #: know globally here.
        self.interface = service_generator.interface
        #: The botocore service model for our service.
        self.service_model = session.get_service_model(self.service_name)
        #: The documentation formatter we will use to format docstrings.
        self.docformatter = DocumentationFormatter()
        #: The classes we've generated for this service.  The key is the class
        #: name, and the value is the code for the class.  We'll write this to a
        #: file later in the order that the classes were generated.
        self.classes: OrderedDict[str, str] = OrderedDict()
        #: A list of imports we need for our classes to function properly.  They'll
        #: be added to the top of the file.
        self.imports: Set[str] = set()

    @property
    def shapes(self) -> List[str]:
        """
        List the names of all the shapes in the service model.

        Returns:
            A list of shape names.

        """
        return self.service_model.shape_names

    def clear(self) -> None:
        """
        Clear the generated classes and imports.
        """
        self.classes = OrderedDict()
        self.imports = set()

    def get_shape(self, name: str) -> botocore.model.Shape:
        """
        Get a :py:class:`botocore.model.Shape` by name from the service model,
        :py:attr:`service_model`.

        Args:
            name: The name of the shape to retrieve.

        Returns:
            The shape object.

        """
        try:
            return self.service_model.shape_for(name)
        except botocore.model.NoShapeFoundError:
            model_name = self.service_def.resolve_model_name(name)
            return self.service_model.shape_for(model_name)

    def get_metadata(self) -> Dict[str, Any]:
        """
        Get the metadata for the botocore service definition.

        Returns:
            The contents of the metadata attribute.

        """
        return self.service_model.metadata

    def generate_all_service_models(self) -> None:
        raise NotImplementedError


class BotocoreFieldsFormatter:
    """
    Handles the formatting of fields from botocore shapes into Python code for
    model classes.  This class extracts and formats field definitions from
    botocore shapes, applying appropriate types, docstrings, and pydantic field
    configurations.
    """

    def __init__(self, model_generator: "ModelGenerator") -> None:
        """
        Initialize the formatter with a reference to the model generator.

        Args:
            model_generator: The ModelGenerator instance that's using this
                formatter.

        """
        #: Reference to the parent ModelGenerator
        self.model_generator = model_generator
        #: The documentation formatter used to format field docstrings
        self.docformatter = self.model_generator.docformatter

    def format_fields(
        self,
        model_name: str,
        model_def: ModelDefinition,
        model_shape: botocore.model.Shape | None = None,
    ) -> List[str]:
        """
        Return the Python code representing all the fields for a model from the
        botocore shape.

        Args:
            model_name: The name of the model to generate fields for
            model_def: The botocraft model definition for this model
            model_shape: The botocore shape to generate the model for. If not
                provided, it will be extracted from the botocore service model.

        Returns:
            The list of Python code representing the pydantic model fields as
            strings

        """
        fields: List[str] = []
        # Get all field definitions for this model
        model_fields = self.model_generator.botocore_shape_field_defs(model_name)

        # Get the shape if not provided
        if not model_shape:
            model_shape = self.model_generator.get_shape(model_name)

        # Use alternate_name if specified
        effective_model_name = model_def.alternate_name or model_name

        # Only proceed if the shape has members
        if not hasattr(model_shape, "members"):
            return fields

        # Process each field in the model
        for field_name, field_def in model_fields.items():
            field_code = self._format_field(
                effective_model_name, field_name, field_def, model_shape
            )
            fields.extend(field_code)

        return fields

    def _format_field(
        self,
        model_name: str,
        field_name: str,
        field_def: ModelAttributeDefinition,
        model_shape: botocore.model.Shape,
    ) -> List[str]:
        """
        Generate code for a single field in a model.

        Args:
            model_name: The model name
            field_name: The field name
            field_def: The field definition
            model_shape: The model shape

        Returns:
            List of lines of code for the field

        """
        # Get field shape and determine if it's required
        field_shape = field_def.botocore_shape
        required = self._is_field_required(field_name, field_def, model_shape)

        # Get type and documentation
        python_type, docstring = self._get_field_type_and_docs(
            model_name, field_name, field_def, field_shape, model_shape
        )

        # Build field definition with appropriate annotations and defaults
        field_code = []
        field_line = self._build_field_definition(
            field_name, field_def, python_type, required
        )
        field_code.append(field_line)

        # Add docstring if available
        if docstring:
            field_code.append(self.docformatter.format_attribute(docstring))

        return field_code

    def _is_field_required(
        self,
        field_name: str,
        field_def: ModelAttributeDefinition,
        model_shape: botocore.model.Shape,
    ) -> bool:
        """
        Determine if a field is required based on field definition and model shape.

        Args:
            field_name: The field name
            field_def: The field definition
            model_shape: The model shape

        Returns:
            True if the field is required, False otherwise

        """
        if field_def.required is not None:
            return field_def.required
        return field_name in model_shape.required_members

    def _get_field_type_and_docs(
        self,
        model_name: str,
        field_name: str,
        field_def: ModelAttributeDefinition,
        field_shape: botocore.model.Shape | None,
        model_shape: botocore.model.Shape,
    ) -> tuple[str, str | None]:
        """
        Get the Python type and documentation for a field.

        Args:
            model_name: The model name
            field_name: The field name
            field_def: The field definition
            field_shape: The field shape
            model_shape: The model shape

        Returns:
            Tuple of (python_type, docstring)

        """
        docstring = field_def.docstring

        if field_shape:
            # Get type from the botocore shape
            python_type = self.model_generator.get_python_type_for_field(
                model_name,
                field_name,
                model_shape=model_shape,
                field_def=field_def,
                field_shape=field_shape,
            )
            # Use shape documentation if no explicit docstring provided
            if not docstring:
                docstring = cast("str", field_shape.documentation)
        else:
            # No shape, must have explicit type in field definition
            assert field_def.python_type, (
                f"Field {field_name} in model {model_name} has no botocore "
                "shape or python type"
            )
            python_type = field_def.python_type

        return python_type, docstring

    def _build_field_definition(
        self,
        field_name: str,
        field_def: ModelAttributeDefinition,
        python_type: str,
        required: bool,
    ) -> str:
        """
        Build the field definition code.

        Args:
            field_name: The field name
            field_def: The field definition
            python_type: The Python type for the field
            required: Whether the field is required

        Returns:
            Field definition code as a string

        """
        # Handle Optional types and default values
        default = None
        if not required:
            if not field_def.readonly and not field_def.rename:
                python_type = f"Optional[{python_type}]"
            default = "None" if field_def.default is None else field_def.default

        # Determine displayed field name (could be renamed)
        display_name = field_def.rename or field_name

        # Start building the field line
        field_line = f"    {display_name}: {python_type}"

        # Determine if we need pydantic Field
        needs_field_class = False
        field_class_args = []

        if default:
            if python_type.startswith("List[") and default == "None":
                field_class_args.append("default_factory=list")
                needs_field_class = True
            elif python_type.startswith("Dict[") and default == "None":
                field_class_args.append("default_factory=dict")
                needs_field_class = True
            else:
                field_class_args.append(f"default={default}")

        if field_def.rename:
            field_class_args.append(f'alias="{field_name}"')
            needs_field_class = True

        if field_def.readonly:
            field_class_args.append("frozen=True")
            needs_field_class = True

        if needs_field_class:
            field_line += f" = Field({', '.join(field_class_args)})"
        elif default:
            field_line += f" = {field_def.default}"

        return field_line


class ExtraFieldsFormatter:
    """
    Handles the formatting of manually defined extra fields into Python code for
    model classes.

    This class is responsible for generating field definitions for extra fields
    that are exclusively defined in the botocraft model definition and not
    present in the botocore service definition.
    """

    def __init__(self, model_generator: "ModelGenerator") -> None:
        """
        Initialize the formatter with a reference to the model generator.

        Args:
            model_generator: The ModelGenerator instance that's using this
                formatter.

        """
        #: Reference to the parent ModelGenerator
        self.model_generator = model_generator
        #: The documentation formatter used to format field docstrings
        self.docformatter = self.model_generator.docformatter

    def format_fields(self, model_def: ModelDefinition) -> List[str]:
        """
        Build out the manually defined extra fields for a model.

        Extra fields are exclusively defined in the botocraft model definition.
        We add them manually to the model definition.

        Args:
            model_def: The botocraft model definition for this model

        Returns:
            A list of formatted field code lines

        """
        fields: List[str] = []

        # Get the extra_fields attribute or return empty list if not present
        extra_fields = getattr(model_def, "extra_fields", {})

        # Process each field
        for field_name, field_def in extra_fields.items():
            field_code = self._format_single_field(field_name, field_def)
            fields.extend(field_code)

            # Update imports for this field
            if hasattr(field_def, "imports") and field_def.imports:
                self.model_generator.imports.update(field_def.imports)

        return fields

    def _format_single_field(
        self, field_name: str, field_def: ModelAttributeDefinition
    ) -> List[str]:
        """
        Format a single extra field into Python code.

        Args:
            field_name: The name of the field
            field_def: The definition of the field

        Returns:
            List of code lines for this field

        """
        field_code = []
        needs_field_class = False
        field_class_args = []

        # Handle default values
        if field_def.default:
            if self._is_container_with_none_default(field_def):
                field_class_args.append(self._get_container_factory(field_def))
                needs_field_class = True
            else:
                field_class_args.append(f"default={field_def.default}")

        # Handle readonly fields
        if field_def.readonly:
            field_class_args.append("frozen=True")
            needs_field_class = True

        # Handle field renaming
        if field_def.rename:
            field_line = f"    {field_def.rename}: {field_def.python_type}"
            field_class_args.append(f'alias="{field_name}"')
            needs_field_class = True
        else:
            field_line = f"    {field_name}: {field_def.python_type}"

        # Append Field class or default value if needed
        if needs_field_class:
            field_line += f" = Field({', '.join(field_class_args)})"
        elif field_def.default:
            field_line += f" = {field_def.default}"

        field_code.append(field_line)

        # Add docstring if available
        if field_def.docstring:
            field_code.append(self.docformatter.format_attribute(field_def.docstring))

        return field_code

    def _is_container_with_none_default(
        self, field_def: ModelAttributeDefinition
    ) -> bool:
        """
        Check if field is a container type (List or Dict) with None default.

        Args:
            field_def: The field definition to check

        Returns:
            True if the field is a container with None default, False otherwise

        """
        return (
            field_def.python_type.startswith("List[")
            or field_def.python_type.startswith("Dict[")
        ) and field_def.default == "None"

    def _get_container_factory(self, field_def: ModelAttributeDefinition) -> str:
        """
        Get the appropriate default_factory for a container type.

        Args:
            field_def: The field definition

        Returns:
            String with the default_factory configuration

        """
        if field_def.python_type.startswith("List["):
            return "default_factory=list"
        if field_def.python_type.startswith("Dict["):
            return "default_factory=dict"
        return f"default={field_def.default}"


class ModelGenerator(AbstractGenerator):
    """
    Generate pydantic model definitions from botocore shapes.
    """

    def __init__(self, service_generator: "ServiceGenerator") -> None:
        super().__init__(service_generator)
        #: The shape converter we will use to convert botocore shapes to python types
        self.shape_converter = PythonTypeShapeConverter(service_generator, self)
        #: The fields formatter for converting botocore shapes to field definitions
        self.fields_formatter = BotocoreFieldsFormatter(self)
        #: The formatter for extra fields defined in the botocraft model definition
        self.extra_fields_formatter = ExtraFieldsFormatter(self)

    def get_model_def(self, model_name: str) -> ModelDefinition:
        """
        Return the :py:class:`ModelDefinition` for a model.

        Notes:
            If there is no human defined model definition for the model, we will
            create a default one.

        Args:
            model_name: The name of the model to get the definition for.

        Returns:
            The model definition.

        """
        if model_name in self.service_def.primary_models:
            defn = self.service_def.primary_models[model_name]
            if defn.readonly:
                defn.base_class = "ReadonlyPrimaryBoto3Model"
            else:
                defn.base_class = "PrimaryBoto3Model"
            return defn
        if model_name in self.service_def.secondary_models:
            defn = self.service_def.secondary_models[model_name]
            if defn.readonly:
                defn.base_class = "ReadonlyBoto3Model"
            else:
                defn.base_class = "Boto3Model"
            return defn
        return ModelDefinition(base_class="Boto3Model", name=model_name)

    def add_extra_fields_from_output_shape(
        self, model_name: str, model_fields: Dict[str, ModelAttributeDefinition]
    ) -> Dict[str, ModelAttributeDefinition]:
        """
        Extract extra fields from the output shape of a get or list method.
        These are fields that are in the output of the get/list methods but not
        in the service models.  We add them to the models as readonly fields so
        that we can load them from the API responses.

        Args:
            model_name: the name of the model to add fields to
            model_fields: the botocraft model field definitions for the model

        Returns:
            ``model_fields`` with any extra fields from the output shape added

        """
        model_def = self.get_model_def(model_name)
        if not model_def.output_shape:
            return model_fields
        output_shape = self.get_shape(model_def.output_shape)
        if not hasattr(output_shape, "members"):
            return model_fields
        for field_name, field_shape in output_shape.members.items():
            if field_name not in model_fields:
                model_fields[field_name] = ModelAttributeDefinition(
                    readonly=True, botocore_shape=field_shape
                )
        return model_fields

    def mark_readonly_fields(
        self, model_name: str, model_fields: Dict[str, ModelAttributeDefinition]
    ) -> Dict[str, ModelAttributeDefinition]:
        """
        Mark model fields as readonly if they are not in any of the input shapes
        defined for the model.  Such fields are returned by AWS but are not
        settable by the user.

        Args:
            model_name: The name of the model to mark fields for.
            model_fields: The botocraft model field definitions for the model.

        Returns:
            The updated model fields.

        """
        model_def = self.get_model_def(model_name)
        if not model_def.input_shapes:
            return model_fields
        # First include any fields that were manually set to writable
        # in the botocraft model definition
        writable_fields: Set[str] = {
            field_name
            for field_name in model_def.fields
            if model_def.fields[field_name].readonly is False
        }
        # Now add any fields that are in the input shapes
        for input_shape_name in model_def.input_shapes:
            input_shape = self.get_shape(input_shape_name)
            if hasattr(input_shape, "members"):
                writable_fields.update(input_shape.members.keys())
        # Mark any fields that are not in writable_fields as readonly
        for field_name, field_def in model_fields.items():
            if field_name not in writable_fields:
                field_def.readonly = True
        return model_fields

    def botocore_shape_field_defs(
        self, model_name: str
    ) -> Dict[str, ModelAttributeDefinition]:
        """
        Return the fields for a botocore shape as a dictionary of field names to
        botocraft field definitions.  This incorporates settings from the
        :py:class:`ModelAttributeDefinitions` from the  ``fields`` attribute of
        the associated model definition, if it exists.

        Note:
            This really only makes sense on
            :py:class:`botocore.model.StructureShape` objects or bespoke
            models, since they are the only ones that have fields (aka
            "members").

        Side Effects:
            If this is not a bespoke model, we set
            :py:attr:`ModelAttributeDefinition.botocore_shape` on each field
            definition, and we set :py:attr:`ModelAttributeDefinition.readonly`
            if the field is not in any of the input shapes for the model.

        Returns:
            A dictionary of field names to field definitions.

        """
        model_def = self.get_model_def(model_name)
        if model_def.bespoke:
            return model_def.extra_fields
        model_shape = self.get_shape(model_name)
        fields: Dict[str, ModelAttributeDefinition] = deepcopy(model_def.fields)
        if hasattr(model_shape, "members"):
            for field, field_shape in model_shape.members.items():
                if field not in fields:
                    fields[field] = ModelAttributeDefinition()
                    if field in model_shape.required_members:
                        fields[field].required = True
                fields[field].botocore_shape = field_shape
        fields = self.mark_readonly_fields(model_name, fields)
        # These are our manually defined extra fields
        if model_def.extra_fields:
            fields.update(model_def.extra_fields)
        # These are the fields that are in the output shape of the get/list
        # methods but not in the service model shape
        fields = self.add_extra_fields_from_output_shape(model_name, fields)
        return fields  # noqa: RET504

    def generate_all_service_models(self) -> None:
        """
        Generate all the service models.
        """
        for model_name in self.service_def.primary_models:
            _ = self.generate_single_model(model_name)

    def format_extra_fields_from_model_def(
        self, model_def: ModelDefinition
    ) -> List[str]:
        """
        Build out the manually defined extra fields for a model.

        Extra fields are exclusively defined in the botocraft model definition.
        We add them manually to the model definition.

        Args:
            model_def: The botocraft model definition for this model

        Returns:
            A list of extra fields.

        """
        return self.extra_fields_formatter.format_fields(model_def)

    def get_properties(self, model_def: ModelDefinition, base_class: str) -> str | None:
        """
        Handle the special properties and methods for primary models.  A primary
        model is a model that has either ``PrimaryBoto3Model`` or
        ``ReadonlyPrimaryBoto3Model`` as its ``base_class``.   Primary models are
        the main models that represent AWS resources, and are those that users can
        create, update, and delete.

        This means:

        * Adding a ``pk`` property that is an alias for the primary key.
        * Maybe adding a ``arn`` property that is an alias for the ARN key.
        * Maybe adding a ``name`` property that is an alias for the name key.
        * Adding any extra properties that were defined in the model definition.
        * Adding any relations to other models that were defined in the model
          definition.
        * Adding any manager shortcut methods that were defined in the model
          definition.

        Args:
            model_def: the botocraft model definition for this model
            base_class: the base class for this model

        Returns:
            The properties for this model, or ``None`` if this is not a primary
            model.

        """
        properties: str = ""
        if base_class in ["PrimaryBoto3Model", "ReadonlyPrimaryBoto3Model"]:
            assert model_def.primary_key or "pk" in model_def.properties, (
                f'Primary service model "{model_def.name}" has no primary key defined'
            )

            if "pk" not in model_def.properties and model_def.primary_key:
                # There is no ``pk`` property, in the ``properties:`` section,
                # but there is a ``primary_key:`` attribute.  We need to add a
                # ``pk`` property that is an alias for the primary key.
                properties = f'''
    @property
    def pk(self) -> Optional[str]:
        """
        Return the primary key of the model.   This is the value of the
        :py:attr:`{model_def.primary_key}` attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.{model_def.primary_key}
'''
            if model_def.arn_key:
                properties += f'''

    @property
    def arn(self) -> Optional[str]:
        """
        Return the ARN of the model.   This is the value of the
        :py:attr:`{model_def.arn_key}` attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.{model_def.arn_key}
'''

            if model_def.name_key:
                properties += f'''

    @property
    def name(self) -> Optional[str]:
        """
        Return the name of the model.   This is the value of the
        :py:attr:`{model_def.name_key}` attribute.

        Returns:
            The name of the model instance.
        """
        return self.{model_def.name_key}
'''

            if model_def.primary_key:
                properties += f'''
    def __hash__(self) -> int:
        """
        Return the hash of the model.   This is the value of the
        :py:attr:`{model_def.primary_key}` attribute.
        """
        return hash(self.{model_def.primary_key})
'''
        # Build any regular properties that were defined in the model definition
        for property_name in model_def.properties:
            if not properties:
                properties = ""
            properties += ModelPropertyGenerator(
                self, model_def.name, property_name
            ).code

        # Now build the relations to other models
        for property_name in model_def.relations:
            if not properties:
                properties = ""
            properties += ModelRelationGenerator(
                self, model_def.name, property_name
            ).code

        # Now build the manager shortcut methods
        for method_name in model_def.manager_methods:
            if not properties:
                properties = ""
            # This acutally needs to be the official name of the model from our
            # botocraft model definition, not the alternate_name if it has one
            properties += ModelManagerMethodGenerator(
                self.service_generator, model_def.name, method_name
            ).code

        return properties

    def get_python_type_for_field(
        self,
        model_name: str,
        field_name: str,
        field_def: ModelAttributeDefinition | None = None,
        model_shape: botocore.model.Shape | None = None,
        field_shape: botocore.model.Shape | None = None,
    ) -> str:
        """
        Return the python type annotation for a field on a model by combining
        the ``model_shape``, ``field_shape`` and ``field_def``.

        Args:
            model_name: The name of the model to get the field from.
            field_name: The name of the field.

        Keyword Args:
            field_def: The botocraft field definition for the field.  If not
                provided, we will look it up in the model definition.
            model_shape: The shape of the model.  If not provided, we will look
                it up in the service model.
            field_shape: The shape of the field.  If not provided, we will look
                it up in the model shape.

        Returns:
            The python type annotation for the field.

        """
        if not model_shape:
            model_shape = self.get_shape(model_name)
        if not hasattr(model_shape, "members"):
            raise ModelHasNoMembersError(model_name)
        if not field_shape:
            field_shape = cast(
                "botocore.model.StructureShape", model_shape
            ).members.get(field_name)
        if not field_def:
            model_def = self.get_model_def(model_name)
            field_def = model_def.fields.get(field_name, ModelAttributeDefinition())
        python_type = field_def.python_type
        if not field_shape and not python_type:
            msg = (
                f"{model_name}.{field_name} has neither a botocore shape nor a "
                "manually defined python_type."
            )
            raise TypeError(msg)
        if not python_type:
            python_type = self.shape_converter.convert(
                cast("botocore.model.StructureShape", field_shape)
            )
        return self.validate_python_type(model_name, field_name, field_def, python_type)

    def validate_python_type(
        self,
        model_name: str,
        field_name: str,
        field_def: ModelAttributeDefinition,
        python_type: str,
    ) -> str:
        """
        After we have guessed the python type for a field, we need to validate
        it to make sure it's not going to cause problems for us later.

        Args:
            model_name: the name of the model
            field_name: the name of the field
            field_def: the botocraft field definition for the field
            python_type: the python type we have guessed for the field

        Raises:
            NoPythonTypeError: we could not determine the type for the field
            TypeError: we have determined that the type for the field is invalid

        Returns:
            The python type annotation for the field on the model.

        """
        if python_type is None:
            raise NoPythonTypeError(field_name, model_name)
        name = field_def.rename if field_def.rename else field_name
        if python_type == name or f"[{name}]" in python_type:
            # If our type annotation is for a model with the same name as the field
            # we'll get recursion errors when trying to import the file.  Quoting
            # the type annotation fixes this sometimes.
            python_type = f'"{python_type}"'
        if field_def.readonly and (
            python_type == f'"{name}"' or f'["{name}"]' in python_type
        ):
            # If the field is readonly, and the type is equal to the field name,
            # even if it is quoted, we will get a "TypeError: forward references
            # must evaluate to types".  This happens because when the field is
            # readonly, we set it equal to ``Field(frozen=True, default=None)``.
            # This causes python typing a lot of consternation, and it throws
            # the TypeError.
            msg = (
                f"Field {model_name}.{name} has type equal to its name, "
                'but is marked as readonly.  This will cause a "TypeError: forward '
                'references must evaluate to types". Fix this in '
                f"botocraft/data/{self.service_generator.aws_service_name}/models.yml ."
                "by either giving an alternate_name for the model named {name} or "
                f"by renaming the field {model_name}.{name} with the rename attribute."
            )
            raise TypeError(msg)
        if not field_def.required and (
            python_type == f'"{name}"' or f'["{name}"]' in python_type
        ):
            # If the field is optional with a None default value, and the type
            # is equal to the field name, pydantic will throw an exception when
            # trying to load data into that field: "ValidationError: Input
            # should be None".  This happens even the type is quoted.
            msg = (
                f"Field {model_name}.{name} has type equal to its name, "
                'but is marked as optional.  This will cause a "ValidationError: Input '
                'should be None" exception from pydantic when trying to load data '
                "into this field.  Fix this in "
                f"botocraft/data/{self.service_generator.aws_service_name}/models.yml ."
                "by either giving an alternate_name for the model named {name} or "
                f"by renaming the field {model_name}.{name} with the rename attribute."
            )
            raise TypeError(msg)

        return python_type

    def format_fields_from_botocore_shape(
        self,
        model_name: str,
        model_def: ModelDefinition,
        model_shape: botocore.model.Shape | None = None,
    ) -> List[str]:
        """
        Return the python code representing all the fields for a model from the
        botocore shape.  This is used to generate the fields for a model when we
        do have a botocore shape for it.

        Args:
            model_name: The name of the model to generate
            model_def: The botocraft model definition for this model

        Keyword Args:
            model_shape: The botocore shape to generate the model for

        Returns:
            The list of python code representing the pydantic model fields as
            strings

        """
        return self.fields_formatter.format_fields(model_name, model_def, model_shape)

    def handle_tag_class(
        self,
        model_def: ModelDefinition,
        base_class: str,
    ) -> ModelTagsDAO:
        """
        Determine whether we need to add
        :py:class:`~botocraft.mixins.tags.TagsDictMixin` as a mixin for the
        model, and if so, what the name of the tag class is.

        If the model has a ``Tags`` or ``TagList`` field case was, we will add
        the :py:model:`~botocraft.mixins.tags.TagsDictMixin` class to the base
        class for this model.  This is used to add the ``tags`` model property
        that is a writable dictionary of tags for the model instead of a list of
        dicts, as AWS uses.

        Args:
            model_def: The botocraft model definition for this model.
            base_class: The current base class(es) for this model.

        Returns:
            The new set of base classes for this model, possibly including the
            ``TagsDictMixin`` class.

        """
        model_name = model_def.name
        if model_def.bespoke:
            # If this is a bespoke model, we need just look at the
            # ModelDefinition.extra_fields for the tags field -- there won't
            # be a botocore shape for this model.   model_fields here are the
            # definitoins for the extra fields in the botocraft model definition.
            model_fields = model_def.extra_fields
        else:
            # This is a botocore model, so we need to get the fields from the
            # botocore model shape.  THis includes the extra fields that are
            # manually defined in the botocraft model definition.  This is
            # a list of model field definitions, augmented from data from the
            # corresponding members of the botocore shappe.
            model_fields = self.botocore_shape_field_defs(model_name)
        field_names = {name.lower(): name for name in model_fields}

        # See if we need to add the TagsDictMixin
        tag_class: str | None = None
        tag_attr: str | None = None
        for name in ["tags", "taglist"]:
            if name in field_names:
                # set the tag_attr to the normally capitalized name of the field
                # in the model definition
                tag_attr = field_names[name]
                break
        if tag_attr:
            # We do have tags defined in the model definition
            if tag_attr != "Tags" and model_fields[tag_attr].rename != "Tags":
                warnings.warn(  # noqa: B028
                    f'Model {model_name} has a field named "tags".  Rename it '
                    'to "Tags" in the model definition.'
                )
            # Extract the name of the tag class from the python type annotation.
            # Different services use different types for tags.
            if model_def.bespoke:
                # If this is a bespoke model, we need to look at the
                # ModelDefinition.extra_fields for the tags field -- there won't
                # be a botocore shape for this model.
                tag_class = model_fields[tag_attr].python_type
                assert tag_class, (
                    f"Bespoke Model {model_name} has a Tag field named "
                    f"{tag_attr} but no python type annotation for it."
                )
            else:
                tag_class = self.get_python_type_for_field(
                    model_name, tag_attr, model_fields[tag_attr]
                )
            if tag_class == "Dict[str, str]":
                # Almost all tags are stored as List[Dict[str, str]] in the
                # botocore model, but some oddballs are actually stored in
                # the exact form we want: Dict[str, str].  In this case, we
                # don't need to add the TagsDictMixin class.
                tag_class = None
            else:
                tag_class = re.sub(r"List\[(.*)\]", r"\1", tag_class)
                tag_class = re.sub(r'"(.*)"', r"\1", tag_class)
                base_class = f"TagsDictMixin, {base_class}"
        return ModelTagsDAO(base_class=base_class, tag_class=tag_class)

    def generate_single_model(  # noqa: PLR0912
        self, model_name: str, model_shape: botocore.model.Shape | None = None
    ) -> str:
        """
        Generate the code for a single model and its dependent models and save
        them to :py:attr:`classes`.

        Args:
            model_name: The name of the model to generate. This will be the
                name of the class.

        Keyword Args:
            model_shape: The botocore shape to generate the model for.  If not provided,
                we will look it up in the service model.

        Side Effects:
            This may add new models to :py:attr:`classes` and new imports to
            :py:attr:`imports`.

        Returns:
            The name of the model class that was generated.

        """
        # Save the original model name so we can use it later when
        # we're looking up the model definition and fields.
        orig_model_name = model_name
        if model_name in self.classes:
            # If we've already generated this model, just return it.
            return model_name

        # The list of Python code lines for this model representing the
        # pydantic model fields.
        field_code: List[str] = []

        # Get the model definition for this model from the service definition
        # file in ``botocraft/data/<service_name>/models.yml``.
        model_def = self.get_model_def(model_name)
        # Handle the case where the model name is different from the
        # botocore model name.
        if model_def.alternate_name:
            model_name = model_def.alternate_name
        # Get the base class for this model
        base_class = cast("str", model_def.base_class)

        if model_def.bespoke:
            # This is not a botocore model, so we just add the code for the
            # fields defined in the ModelDefinition.extra_fields attribute.
            field_code = self.format_extra_fields_from_model_def(model_def)
        else:
            # This is a botocore model, so we generate the code for fields from a
            # combination of the botocore model shape and the botocraft model
            # definition, including any extra fields that are defined in the
            # botocraft model definition.
            model_shape = self.get_shape(orig_model_name)
            field_code = self.format_fields_from_botocore_shape(
                orig_model_name, model_def, model_shape=model_shape
            )

        # Add any botocraft defined properties.  This includes relations.
        properties = self.get_properties(model_def, base_class)

        # Add any botocraft defined mixins to the class inheritance
        if model_def.mixins:
            for mixin in model_def.mixins:
                self.imports.add(f"from {mixin.import_path} import {mixin.name}")
            base_class = ", ".join(
                [mixin.name for mixin in model_def.mixins] + [base_class]
            )

        tags_dao = self.handle_tag_class(model_def, base_class)
        code: str = f"class {model_name}({tags_dao.base_class}):\n"
        if model_shape:
            # If we have a botocore shape for this model, get the docstring
            # for the class from the shape.
            docstring = self.docformatter.format_docstring(model_shape.documentation)
        else:
            # If we don't have a botocore shape for this model, get the
            # docstring from the botocraft model definition, if it exists.
            _docstring = model_def.docstring if model_def.docstring else None
            if _docstring:
                docstring = self.docformatter.format_docstring(_docstring)
            else:
                # If we don't have a docstring, we need to set it to None
                # so that the docformatter doesn't try to format it.
                docstring = None
        if docstring:
            code += f'    """{docstring}"""\n'
        if tags_dao.tag_class:
            # The ``tag_class`` attribute is the class that will be used to
            # represent the tags in the model.  This is used by TagsDictMixin
            # to convert the tags to a dictionary of tag key/value pairs.
            code += f"    tag_class: ClassVar[Type] = {tags_dao.tag_class}\n"
        if "PrimaryBoto3Model" in base_class:
            if model_def.alternate_name:
                manager_name = f"{model_def.alternate_name}Manager"
            else:
                manager_name = f"{model_name}Manager"
            code += f"    manager_class: ClassVar[Type[Boto3ModelManager]] = {manager_name}\n\n"  # noqa: E501
        if field_code:
            code += "\n".join(field_code)
        if properties:
            code += f"\n{properties}"
        if not field_code and not properties:
            code += "    pass"
        # Add the class to the list so we only generate it once and can
        # write it out to the file later.
        self.classes[model_name] = code
        assert code, (
            f"Model {model_name} has no fields or properties defined.  This is "
            "probably a bug in the botocraft model definition or generator code."
        )
        return model_name


class ManagerGenerator(AbstractGenerator):
    """
    Generates the code for the manager class for a service.

    Args:
        service_generator: The :py:class:`ServiceGenerator` we're generating
            models for.

    """

    #: A mapping of botocore operation names to the method generator class that
    #: will generate the code for that method.
    METHOD_GENERATORS: Final[Dict[str, Type[ManagerMethodGenerator]]] = {
        "create": CreateMethodGenerator,
        "update": UpdateMethodGenerator,
        "partial_update": PartialUpdateMethodGenerator,
        "delete": DeleteMethodGenerator,
        "get": GetMethodGenerator,
        "get_many": GetManyMethodGenerator,
        "list": ListMethodGenerator,
    }

    def __init__(self, service_generator: "ServiceGenerator") -> None:
        super().__init__(service_generator)
        self.model_generator = self.service_generator.model_generator
        self.shape_converter = self.model_generator.shape_converter
        self.client = boto3.client(self.service_name)  # type: ignore[call-overload]

    def generate_manager(self, model_name: str, manager_def: ManagerDefinition) -> None:
        """
        Generate the code for a single manager, and its dependent response
        classes and save them to :py:attr:`classes`.

        Args:
            model_name: The name of the model to generate the manager for.
            manager_def: The botocraft manager definition for the manager.

        """
        methods: OrderedDict[str, str] = OrderedDict()
        for method_name, method_def in manager_def.methods.items():
            generator = self.get_method_generator(model_name, method_name, method_def)
            methods[method_name] = generator.code
        method_code = "\n\n".join(methods.values())
        base_class = "Boto3ModelManager"
        model_def = self.model_generator.get_model_def(model_name)
        if model_def.alternate_name:
            manager_name = f"{model_def.alternate_name}Manager"
        else:
            manager_name = f"{model_name}Manager"

        # Add any botocraft defined mixins to the class inheritance
        if manager_def.mixins:
            for mixin in manager_def.mixins:
                self.imports.add(f"from {mixin.import_path} import {mixin.name}")
            base_class = ", ".join(
                [mixin.name for mixin in manager_def.mixins] + [base_class]
            )

        # If this is a readonly manager, we need to use the readonly manager
        # base class
        if manager_def.readonly:
            base_class = "ReadonlyBoto3ModelManager"
        code = f"""


class {manager_name}({base_class}):

    service_name: str = '{self.service_name}'

{method_code}
"""
        self.classes[manager_name] = code

    def get_method_generator(
        self, model_name: str, method_name: str, method_def: ManagerMethodDefinition
    ) -> ManagerMethodGenerator:
        """
        Return the appropriate method generator class for a given method
        definition.

        Args:
            model_name: the model name for the manager we're generating
            method_name: the name of the method we're generating
            method_def: the method definition for the method we're generating

        Returns:
            A method generator class.

        """
        try:
            method_generator_class = self.METHOD_GENERATORS[method_name]
        except KeyError:
            # We have no specific method generator for this method, so we
            # will use the general method generator.
            generator: ManagerMethodGenerator = GeneralMethodGenerator(
                self, model_name, method_def, method_name=method_name
            )
        else:
            # We have a specific method generator for this method, so we
            # will use that.
            generator = method_generator_class(self, model_name, method_def)

        return generator

    def generate_all_manager_models(self) -> None:
        for model_name, manager_def in self.service_def.managers.items():
            self.generate_manager(model_name, manager_def)
        self.imports.update(self.model_generator.imports)


class ServiceGenerator:
    """
    Generate the code for a single AWS service.

    This means:

        * Managers
        * Service Models
        * Request/Response Models

    Args:
        service_def: The :py:class:`ServiceDefinition` for the service we are
            generating code for.

    """

    service_path: Path = Path(__file__).parent.parent / "services"

    def __init__(self, service_def: ServiceDefinition) -> None:
        #: The service definition
        self.service_def = service_def
        #: The botocraft interface object, where we will collect all our global data
        self.interface = service_def.interface
        #: A set of model imports we need to add to the top of the file
        self.imports: Set[str] = {
            "from datetime import datetime",
            "from typing import ClassVar, Type, Optional, Literal, Dict, List, Union, Any, cast",  # noqa: E501
            "from pydantic import Field",
            "from .abstract import Boto3Model, ReadonlyBoto3Model, PrimaryBoto3Model, "
            "ReadonlyPrimaryBoto3Model, Boto3ModelManager, ReadonlyBoto3ModelManager",
            "from botocraft.mixins.tags import TagsDictMixin",
        }
        #: A dictionary of model names to class code.  This is populated by
        #: service models
        self.model_classes: Dict[str, str] = {}
        #: A dictionary of botocore response classes names to class code. This
        #: is populated when we build the manager classes
        self.response_classes: Dict[str, str] = {}
        #: A dictionary of manager classes names to class code. This is populated
        #: when we build the manager classes
        self.manager_classes: Dict[str, str] = {}
        #: The :py:class:`ModelGenerator` class we will use to generate models
        self.model_generator = ModelGenerator(self)
        #: The :py:class:`ManagerGenerator` class we will use to generate managers
        self.manager_generator = ManagerGenerator(self)
        #: The :py:class:`ManagerGenerator` class we will use to generate managers
        self.sphinx_generator = ServiceSphinxDocBuilder(self)

    @property
    def aws_service_name(self) -> str:
        """
        Return the boto3 service name for this service.
        """
        return self.service_def.name

    @property
    def safe_service_name(self) -> str:
        """
        Return the safe service name for this service.  This is the name we
        use to generate the file name for the service.
        """
        return self.service_def.safe_service_name

    @property
    def service_full_name(self) -> str:
        """
        Return what AWS thinks the full name of this services is.

        Returns:
            The full name of the service, as defined in the botocore service.

        """
        return self.model_generator.service_model.metadata["serviceId"]

    @property
    def classes(self) -> Dict[str, str]:
        """
        Return a dictionary of all the classes we have generated.
        """
        return {
            **self.model_classes,
            **self.response_classes,
            **self.manager_classes,
        }

    @property
    def code(self) -> str:
        """
        The code for this service.
        """
        imports = "\n".join(list(self.imports))
        model_classes = "\n\n".join(self.model_classes.values())
        response_classes = "\n\n".join(self.response_classes.values())
        manager_classes = "\n\n".join(self.manager_classes.values())
        return f"""
# This file is automatically generated by botocraft.  Do not edit directly.
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
{imports}

# ===============
# Managers
# ===============

{manager_classes}


# ==============
# Service Models
# ==============

{model_classes}


# =======================
# Request/Response Models
# =======================

{response_classes}

"""

    def generate(self) -> None:
        """
        Generate the code for this service.
        """
        # Generate the service models
        self.model_generator.generate_all_service_models()
        self.model_classes = deepcopy(self.model_generator.classes)
        self.imports.update(self.model_generator.imports)

        # Generate the service managers and request/response models
        self.manager_generator.generate_all_manager_models()
        # We have to do this because very occasionaly there are no real
        # primary models, and we use a response model as one.  Thus we need
        # to remove any primary models from the response classes
        self.response_classes = {
            k: v
            for k, v in self.model_generator.classes.items()
            if k not in self.model_classes
        }
        self.manager_classes = deepcopy(self.manager_generator.classes)
        self.imports.update(self.manager_generator.imports)
        self.model_generator.clear()
        self.manager_generator.clear()

        # Write the generated code to the output file
        self.write()
        # Write the sphinx documentation for the service
        self.sphinx_generator.write()

        # Update the interface with the manager models we generated
        for model_name in self.model_classes:
            self.interface.add_model(model_name, self.service_def.name)
        for model_name in self.response_classes:
            self.interface.add_model(model_name, self.service_def.name)
        for model_name in self.manager_classes:
            self.interface.add_model(model_name, self.service_def.name)

    def write(self) -> None:
        """
        Write the generated code to the output file, and format it with ruff for
        code formatting and auto-fixing, then docformatter for docstring formatting.

        """
        code = self.code
        output_file = self.service_path / f"{self.service_def.safe_service_name}.py"

        # Use ruff to format and auto-fix the code
        with tempfile.TemporaryDirectory(dir=self.service_path) as temp_dir:
            temp_path = Path(temp_dir) / f"{self.service_def.safe_service_name}.py"
            init_path = Path(temp_dir) / "__init__.py"
            with init_path.open("w", encoding="utf-8") as f:
                f.write(
                    '"""This file is automatically generated by botocraft.  Do not edit directly."""\n'  # noqa: E501
                )

            with temp_path.open("w", encoding="utf-8") as f:
                f.write(code)

            try:
                # Format with ruff (replaces black)
                subprocess.run(
                    ["ruff", "format", str(temp_path)],
                    check=True,
                    capture_output=True,
                    text=True,
                )

                # Fix with ruff (auto-fixes linting issues, including import sorting)
                subprocess.run(
                    ["ruff", "check", "--fix", str(temp_path)],
                    check=True,
                    capture_output=True,
                    text=True,
                )

                # Read the formatted content
                with temp_path.open("r", encoding="utf-8") as f:
                    formatted_code = f.read()

            except subprocess.CalledProcessError as e:
                # Print the problematic code with line numbers for easier debugging
                code_lines = [
                    f"{i:04} " + line for i, line in enumerate(code.split("\n"))
                ]
                print("\n".join(code_lines))
                print(f"ruff error: {e.stdout}\n{e.stderr}")
                raise

        # Format the docstrings with docformatter
        formatted_code = Formatter(FormatterArgs(), None, None, None)._format_code(  # noqa: SLF001
            formatted_code
        )

        # Write the final formatted code to the output file
        with output_file.open("w", encoding="utf-8") as fd:
            fd.write(formatted_code)



from functools import cached_property
from typing import TYPE_CHECKING, Any, Dict, Type, cast

if TYPE_CHECKING:
    from botocraft.services.abstract import Boto3Model


class TagsDict(dict):
    """
    A dict subclass that allows python dictionary style access to the
    ``Tags`` attribute for a model instance.   Normally ``Tags`` is a list of
    dictionaries, but this class allows you to treat tags like a dictionary
    where the keys are the tag names and the values are the tag values.

    This subclass does the following additional things:

    * overrides the ``__setitem__`` method to set the the appropriate key
      and value in the ``Tags`` attribute of the model instance.
    * overrides the ``__delitem__`` method to delete the appropriate key
      and value in the ``Tags`` attribute of the model instance.
    * overrides the ``pop`` method to delete the appropriate key
      and value in the ``Tags`` attribute of the model instance.
    """

    def __init__(self, *args, **kwargs) -> None:
        #: The model instance that we are setting tags for.
        self.instance: "Boto3Model" | None = None  # noqa: UP037
        #: The boto3 tag class.  This has to be configurable because once
        #: again boto3 has different tag classes for different services.
        self.tag_class: Type | None = None
        #: The name of the key attribute in the tag class.
        self.tag_Key: str | None = None
        #: The name of the value attribute in the tag class.
        self.tag_Value: str | None = None
        super().__init__(*args, **kwargs)

    def check(self) -> None:
        """
        Check if we have been configured properly.

        This means:

        * the ``instance`` attribute must be set.
        * ``self.instance`` must have an attribute named ``Tags``.
        * the ``tag_class`` attribute must be set.
        * the ``tag_Key`` attribute must be set.
        * the ``tag_Value`` attribute must be set.

        Returns:
            True if we have been configured, False otherwise.

        """
        assert self.instance is not None, (
            "The instance attribute must be set before setting tags."
        )
        assert hasattr(self.instance, "Tags"), (
            f"The {self.instance.__class__.__name__} does not have a Tags attribute."
        )
        assert self.tag_class is not None, (
            "The tag_class attribute must be set before setting tags."
        )
        assert self.tag_Key is not None, (
            "The tag_Key attribute must be set before setting tags."
        )
        assert self.tag_Value is not None, (
            "The tag_Value attribute must be set before setting tags."
        )

    def __setitem__(self, key: str, value: str) -> None:
        """
        Set the tags in ourselves and in the ``Tags`` attribute of the model
        instance.

        Args:
            key: the key to set.
            value: the value to set.

        """
        self.check()
        found: bool = False
        instance = cast("Boto3Model", self.instance)
        if instance.Tags is None:  # type: ignore[attr-defined]
            instance.Tags = []  # type: ignore[attr-defined]
        for tag in instance.Tags:  # type: ignore[attr-defined]
            if getattr(tag, cast("str", self.tag_Key)) == key:
                setattr(tag, cast("str", self.tag_Value), value)
                found = True
                break
        if not found:
            tag = cast("Type", self.tag_class)(
                **{self.tag_Key: key, self.tag_Value: value}
            )
            instance.Tags.append(tag)  # type: ignore[attr-defined]
        super().__setitem__(key, value)

    def __delitem__(self, key: str) -> None:
        """
        Delete the tags in ourselves and in the ``Tags`` attribute of the model
        instance.

        Args:
            key: the key to delete.

        """
        self.check()
        instance = cast("Boto3Model", self.instance)
        for tag in instance.Tags:  # type: ignore[attr-defined]
            if getattr(tag, cast("str", self.tag_Key)) == key:
                instance.Tags.remove(tag)  # type: ignore[attr-defined]
                break
        super().__delitem__(key)

    def pop(self, key: str, default: Any | None = None) -> Any:
        """
        Override the ``pop`` method to delete the tags in ourselves and in the
        ``Tags`` attribute of the model instance.

        Args:
            key: the key to delete.

        Keyword Args:
            default: the default value to return if the key is not found.

        Returns:
            Returns the value of the key if it exists, otherwise the default

        """
        self.check()
        instance = cast("Boto3Model", self.instance)
        for tag in instance.Tags:  # type: ignore[attr-defined]
            if getattr(tag, cast("str", self.tag_Key)) == key:
                instance.Tags.remove(tag)  # type: ignore[attr-defined]
                break
        return super().pop(key, default)


class TagsDictMixin:
    """
    A mixin that adds a ``tags`` property to a model instance.  This
    is intended to be used with :py:class:`Boto3Model` classes that have
    a ``Tags`` attribute.

    The attribute that has the AWS tags list **must** be named ``Tags``.  If
    your model stores them in another attribute (for example, ECS objects use
    ``tags`` instead), you will need to override the attribute name by setting
    the :py:attr:`botocraft.sync.models.ModelAttributeDefinition.rename`
    attribute to be ``Tags`` for that attribute.

    .. important::
        One important assumption this mixin makes is the caller modifies tags
        via either self.tags or self.Tags, but not both at the same time.  If
        you do use both at the same time, you will get unexpected results.
    """

    #: The boto3 tag class.  This has to be configurable because once
    #: again boto3 has different tag classes for different services.
    tag_class: Type | None = None

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.tag_class is None:
            from botocraft.services.common import Tag

            self.tag_class = Tag

    @cached_property
    def __tag_Key(self) -> str:
        """
        The name of the key attribute in the tag class.
        """
        assert self.tag_class is not None, (
            "The tag_class class attribute must be set before using tags."
        )
        instance = self.tag_class()
        candidates = ["key", "Key"]
        for candidate in candidates:
            if hasattr(instance, candidate):
                return candidate
        msg = f"Could not find a key attribute in {self.tag_class.__name__}"
        raise RuntimeError(msg)

    @cached_property
    def __tag_Value(self) -> str:
        """
        The name of the value attribute in the tag class.
        """
        assert self.tag_class is not None, (
            "The tag_class class attribute must be set before using tags."
        )
        instance = self.tag_class()
        candidates = ["value", "Value"]
        for candidate in candidates:
            if hasattr(instance, candidate):
                return candidate
        msg = f"Could not find a value attribute in {self.tag_class.__name__}"
        raise RuntimeError(msg)

    @property
    def tags(self) -> TagsDict:
        """
        Get the tags for the model instance.

        Returns:
            The tags for the model instance.

        """
        assert hasattr(self, "Tags"), (
            f"The {self.__class__.__name__} does not have a Tags attribute."
        )
        if self.Tags is None:
            _tags = TagsDict()
        else:
            _tags = TagsDict(
                **{
                    getattr(tag, self.__tag_Key): getattr(tag, self.__tag_Value)
                    for tag in self.Tags
                }
            )
        _tags.instance = cast("Boto3Model", self)
        _tags.tag_class = self.tag_class
        _tags.tag_Key = self.__tag_Key
        _tags.tag_Value = self.__tag_Value
        return _tags

    @tags.setter
    def tags(self, value: Dict[str, str] | TagsDict) -> None:
        """
        Set the tags for the model instance.

        Args:
            value: the tags to set for the model instance.

        """
        assert self.tag_class is not None, (
            "The tag_class class attribute must be set before setting tags."
        )
        assert hasattr(self, "Tags"), (
            f"The {self.__class__.__name__} does not have a Tags attribute."
        )
        self.Tags = []
        for key, _value in value.items():
            tag = self.tag_class(**{self.__tag_Key: key, self.__tag_Value: _value})
            self.Tags.append(tag)



# mypy: disable-error-code="attr-defined"

import time
from collections import OrderedDict
from typing import TYPE_CHECKING, ClassVar

import boto3
from botocore.exceptions import WaiterError

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.services import (
        AutoScalingGroupManager,
    )


class AutoScalingGroupModelMixin:
    """
    A mixin for :py:class:`AutoScalingGroup` that adds some convenience methods.

    Sometimes we like full :py:class:`Instance` objects instead of the
    :py:class:`AutoScalingInstanceReference` objects that get listed on
    :py:attr:`AutoScalingGroup.Instances`.

    EC2 is weird and doesn't have an easy way to list instances in an
    autoscaling group except to use one of the ``Filter`` parameters on
    ``describe_instances``.  Instances that are part of an autoscaling group
    have a tag called ``aws:autoscaling:groupName`` whose value is the name of
    the autoscaling group.  We can use this to filter instances that belong to a
    particular autoscaling group.

    This method is too specialized at the moment to be included as one of the
    transformers for model related objects, so we'll just add it as a mixin.
    """

    objects: ClassVar["AutoScalingGroupManager"]

    session: boto3.session.Session
    AutoScalingGroupName: str
    MinSize: int
    MaxSize: int

    @property
    def ec2_instances(self) -> PrimaryBoto3ModelQuerySet:
        """
        Return the running :py:class:`Instance` objects that belong to this
        group, if any.
        """
        # Avoid circular import
        from botocraft.services.ec2 import Instance

        if self.AutoScalingGroupName:
            pk = OrderedDict(
                Filters=[
                    {
                        "Name": "tag:aws:autoscaling:groupName",
                        "Values": [self.AutoScalingGroupName],
                    },
                    {"Name": "instance-state-name", "Values": ["running"]},
                ]
            )
            return Instance.objects.using(self.session).list(**pk)  # type: ignore[attr-defined]
        return PrimaryBoto3ModelQuerySet([])

    @property
    def is_stable(self) -> bool:
        """
        Return ``True`` if the autoscaling group is stable, ``False`` otherwise.

        An autoscaling group is considered stable if all instances are running
        and healthy and the DesiredCapacity is equal to the number of instances.
        """
        ec2_instances = self.ec2_instances
        if len(ec2_instances) == self.DesiredCapacity:  # type: ignore[attr-defined]
            # check if all instances are in service
            instance_ids = [instance.InstanceId for instance in ec2_instances]
            details = self.objects.using(self.session).instance_status(
                InstanceIds=instance_ids
            )
            if all(detail.HealthStatus == "HEALTHY" for detail in details):
                return True
        return False

    def wait_until_stable(self, max_attempts: int = 40, delay: int = 15) -> None:
        """
        Since there is no waiter for this, we'll use this method to wait until
        the autoscaling group is stable.

        Raises:
            botocore.exceptions.WaiterError: If the autoscaling group is not
                stable after ``max_attempts``.

        Keyword Args:
            max_attempts: The maximum number of attempts to make before giving
                up.
            delay: The number of seconds to wait between attempts.

        """
        from botocraft.services import AutoScalingGroupsType

        wait_count: int = 0
        # There is no waiter for this, so we'll just poll until the desired
        # count is reached, or we reach max_attempts.
        while True:
            asg = self.objects.using(self.session).get(
                AutoScalingGroupName=self.AutoScalingGroupName
            )
            assert asg is not None, (
                "AutoScalingGroup.wait_until_stable(): Autoscaling group not found."
            )
            if wait_count >= max_attempts:
                reason = "Max attempts exceeded"
                raise WaiterError(
                    name="asg_stable",
                    reason=reason,
                    last_response=AutoScalingGroupsType(
                        AutoScalingGroups=[asg]
                    ).model_dump(),  # type: ignore[arg-type]
                )
            if asg.is_stable:
                break
            wait_count += 1
            time.sleep(delay)

    def scale(
        self,
        desired_count: int,
        wait: bool = False,
        max_attempts: int = 40,
    ) -> None:
        """
        Scale the autoscaling group to the desired count.

        Args:
            desired_count: The number of tasks to run.

        Keyword Args:
            wait: If True, wait for the service to reach the desired count.
            max_attempts: The maximum number of attempts to make before giving up

        """
        if desired_count < self.MinSize:
            msg = (
                f"desired_count must be greater than or equal to MinSize, "
                f"which is {self.MinSize}."
            )
            raise ValueError(msg)
        if desired_count > self.MaxSize:
            msg = (
                f"desired_count must be less than or equal to MaxSize, "
                f"which is {self.MaxSize}."
            )
            raise ValueError(msg)
        self.objects.using(self.session).scale(self.AutoScalingGroupName, desired_count)
        time.sleep(10)
        if wait:
            wait_count: int = 0
            # There is no waiter for this, so we'll just poll until the desired
            # count is reached, or we reach max_attempts.
            while True:
                if wait_count >= max_attempts:
                    msg = (
                        f"Reached max attempts of {max_attempts} to reach desired "
                        f"count of {desired_count}."
                    )
                    raise TimeoutError(msg)
                if self.is_stable:
                    break
                wait_count += 1
                time.sleep(5)



from __future__ import annotations

from datetime import datetime  # noqa: TC003
from typing import List

from pydantic import BaseModel, Field


class ECRReplicationAction(BaseModel):
    """
    Represents details about an ECR image replication action.  Contains
    information about images being replicated between ECR repositories across
    regions or accounts, along with the result status of the operation.
    """

    #: The type of replication action performed (e.g., "PUSH", "PULL")
    action_type: str = Field(..., alias="action-type")
    #: The SHA256 digest of the image being replicated
    image_digest: str = Field(..., alias="image-digest")
    #: The tag associated with the image being replicated
    image_tag: str = Field(..., alias="image-tag")
    #: The name of the ECR repository where the image is being replicated
    repository_name: str = Field(..., alias="repository-name")
    #: The result of the replication action (e.g., "SUCCESS", "FAILURE")
    result: str
    #: The AWS account ID of the source repository
    source_account: str = Field(..., alias="source-account")
    #: The AWS region of the source repository
    source_region: str = Field(..., alias="source-region")


class ECRReplicationActionEvent(BaseModel):
    """
    Represents a complete ECR Replication Action event from EventBridge.
    Contains metadata about the event and the actual replication action details.
    """

    #: The detailed information about the ECR replication action
    detail: ECRReplicationAction
    #: The AWS account ID where the event occurred
    account: str
    #: The human-readable type of the event (e.g., "ECR Image Replication")
    detail_type: str = Field(..., alias="detail-type")
    #: The unique identifier for this event instance
    id: str
    #: The AWS region where the event occurred
    region: str
    #: The AWS resources involved in this event, typically the repository ARN
    resources: List[str]
    #: The source of the event (typically "aws.ecr")
    source: str
    #: The timestamp when the event was generated, in UTC
    time: datetime
    #: The version of the event schema format
    version: str
